import datetime
from typing import List
from .action import RepoAction
from .config import RepoConfig


class Syncrhonize(RepoAction):

    def __init__(self, config: RepoConfig, graph_files: List[str]) -> None:
        super(
            Syncrhonize,
            self
        ).__init__(config, graph_files)

    def execute(self):

        self.logger.hedra.sync.debug(f'SyncrhonizeAction: {self.action_id} - Re-initializing repo')
        self._setup()

        self.logger.hedra.sync.debug(f'SyncrhonizeAction: {self.action_id} - Fetchin latest branches from remote')
        self.remote.fetch()

        self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Checking out branch - {self.config.branch} - and updating remote tracking')
        self._checkout()

        current_time = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        pre_sync_message = f"Hedra graph update: {self.config.path} - {current_time}" 

        self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Running pre-pull commit.')
        self.repo.git.add(A=True)
        self.repo.index.commit(pre_sync_message)

        self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Pulling latest from remote repo at - URI: {self.config.uri} - Branch: {self.config.branch}')
        self.remote.pull(self.branch.name, rebase=True)

        self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Updating .gitignore')
        self._update_ignore(force_create=True)

        self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Adding {len(self.discovered_files)} to repo in commit')
        self.repo.index.add(self.discovered_files)

        sync_message = self.config.sync_message

        if sync_message is None:
            current_time = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
            sync_message = f"Hedra graph update: {self.config.path} - {current_time}" 

        self.repo.index.commit(sync_message)

        self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Pushing project changes and updates to remote repo at - URI: {self.config.uri} - Branch: {self.config.branch}')
        self.remote.push(self.config.branch, set_upstream=True)





from broadkast.grpc.proto import ServiceDiscoveryResponse
from hedra.runners.bootstrap_services.proto import ServiceRegistrationStatusResponse
from hedra.runners.leader_services.leader_registry.leader_service import LeaderService

class DiscoveredLeadersRegistry:

    def __init__(self) -> None:
        self.count = 0
        self.discovered = {}
        self.ready = False

    async def register(self, registration_response: ServiceRegistrationStatusResponse, host: str=None):
        self.count: int = registration_response.registered_count

        for node_address, node_id in zip(registration_response.registered_addresses, registration_response.registered_ids):
            self.discovered[node_address] = LeaderService(
                node_address,
                node_id,
                host=host
            )

        self.ready = registration_response.ready

    async def register_kubernetes(self, discovered: ServiceDiscoveryResponse, host: str=None):
        self.count: int = discovered.discovered_count
        
        for instance in discovered.instances:
            registered_address = instance.instance_address
            self.discovered[registered_address] = LeaderService(
                instance.instance_address, 
                instance.instance_id, 
                host=instance.instaregistrationnce_address == host
            )
            
        self.ready = True

        


import asyncio
import functools
import psutil
import uuid
from typing import List
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    import libhoney
    from .honeycomb_config import HoneycombConfig
    has_connector = True

except Exception:
    libhoney = None
    HoneycombConfig = None
    has_connector = False


class Honeycomb:

    def __init__(self, config: HoneycombConfig) -> None:
        self.api_key = config.api_key
        self.dataset = config.dataset
        self.client = None
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=True))
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()


    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Honeycomb.IO')
        self.client = await self._loop.run_in_executor(
            None,
            functools.partial(
                libhoney.init,
                writekey=self.api_key,
                dataset=self.dataset
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Honeycomb.IO')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Honeycomb.IO')

        for event in events:

            Honeycomb_event = libhoney.Event(data={
                **event.record
            })

            await self._loop.run_in_executor(
                self._executor,
                Honeycomb_event.send
            )

        await self._loop.run_in_executor(
            self._executor,
            libhoney.flush
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Honeycomb.IO')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Honeycomb.IO')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            group_metric = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                **metrics_set.common_stats
            }

            honeycomb_group_metric = libhoney.Event(data=group_metric)

            await self._loop.run_in_executor(
                self._executor,
                honeycomb_group_metric.send
            )

        await self._loop.run_in_executor(
            self._executor,
            libhoney.flush
        )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Honeycomb.IO')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Honeycomb.IO')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():

                metric_record = {
                    **group.stats, 
                    **group.custom,
                    'group': group_name
                }     
    
                honeycomb_event = libhoney.Event(data=metric_record)

                await self._loop.run_in_executor(
                    self._executor,
                    honeycomb_event.send
                )

        await self._loop.run_in_executor(
            self._executor,
            libhoney.flush
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Honeycomb.IO')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Honeycomb.IO')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():

                metric_record ={
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': custom_group_name,
                    **group
                }

                honeycomb_event = libhoney.Event(data=metric_record)

                await self._loop.run_in_executor(
                    self._executor,
                    honeycomb_event.send
                )

        await self._loop.run_in_executor(
            self._executor,
            libhoney.flush
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Honeycomb.IO')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Honeycomb.IO')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:

                error_event = libhoney.Event(data={
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                })

                await self._loop.run_in_executor(
                    self._executor,
                    error_event.send
                )


        await self._loop.run_in_executor(
            self._executor,
            libhoney.flush
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Honeycomb.IO')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection Honeycomb.IO')

        await self._loop.run_in_executor(
            self._executor,
            libhoney.close
        )

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection Honeycomb.IO')




from .transitions import (
    checkpoint_to_setup_transition,
    checkpoint_to_optimize_transition,
    checkpoint_to_execute_transition,
    checkpoint_to_teardown_transition,
    checkpoint_to_analyze_transition,
    checkpoint_to_complete_transition,
    checkpoint_to_submit_transition,
    checkpoint_to_wait_transition,
    checkpoint_to_checkpoint_transition
)


from .common import ReporterTypes

from .aws_lambda import (
    AWSLambda,
    AWSLambdaConfig
)

from .aws_timestream import (
    AWSTimestream,
    AWSTimestreamConfig
)

from .bigquery import (
    BigQuery,
    BigQueryConfig
)

from .bigtable import (
    BigTable,
    BigTableConfig
)

from .cassandra import (
    Cassandra,
    CassandraConfig
)

from .cloudwatch import (
    Cloudwatch,
    CloudwatchConfig
)

from .cosmosdb import (
    CosmosDB,
    CosmosDBConfig
)

from .csv import (
    CSV,
    CSVConfig
)

from .datadog import (
    Datadog,
    DatadogConfig
)

from .dogstatsd import (
    DogStatsD,
    DogStatsDConfig
)

from .google_cloud_storage import (
    GoogleCloudStorage,
    GoogleCloudStorageConfig
)

from .graphite import (
    Graphite,
    GraphiteConfig
)

from .honeycomb import (
    Honeycomb,
    HoneycombConfig
)

from .influxdb import (
    InfluxDB,
    InfluxDBConfig
)

from .json import (
    JSON,
    JSONConfig
)

from .kafka import (
    Kafka,
    KafkaConfig
)

from .mongodb import (
    MongoDB,
    MongoDBConfig
)

from .mysql import (
    MySQL,
    MySQLConfig
)

from .netdata import (
    Netdata,
    NetdataConfig
)

from .newrelic import (
    NewRelic,
    NewRelicConfig
)

from .postgres import (
    Postgres,
    PostgresConfig
)

from .prometheus import (
    Prometheus,
    PrometheusConfig
)

from .redis import (
    Redis,
    RedisConfig
)

from .s3 import (
    S3,
    S3Config
)

from .snowflake import (
    Snowflake,
    SnowflakeConfig
)

from .sqlite import (
    SQLite,
    SQLiteConfig
)

from .statsd import (
    StatsD,
    StatsDConfig
)

from .telegraf import (
    Telegraf,
    TelegrafConfig
)

from .telegraf_statsd import (
    TelegrafStatsD,
    TelegrafStatsDConfig
)

from .timescaledb import (
    TimescaleDB,
    TimescaleDBConfig
)


from hedra.core.engines.types.grpc import GRPCResult
from .http2_event import HTTP2Event


class GRPCEvent(HTTP2Event):

    def __init__(self, result: GRPCResult) -> None:
        super().__init__(result)


import traceback
import uuid
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.reporting.tags import Tag


class BaseEvent:

    __slots__ = (
        'event_id',
        'action_id',
        'name',
        'shortname',
        'error',
        'type',
        'source',
        'checks',
        'tags',
        'stage',
        'timings',
        'time'
    )

    def __init__(self, result: BaseResult) -> None:

        self.event_id = str(uuid.uuid4())
        self.action_id = result.action_id

        self.name = None
        self.shortname = result.name
        self.error = result.error
        self.timings = {}
        self.type = result.type
        self.source = result.source
        self.checks = []

        self.time = self.timings.get('total', 0)

        for check_hook_name in result.checks:
            check_hook = registrar.all.get(check_hook_name)
            self.checks.append(check_hook.call)

        self.tags = [
            Tag(
                tag.get('name'),
                tag.get('value')
            ) for tag in result.tags
        ]
        self.stage = None

    async def check_result(self):
        if self.error is None:
            for check in self.checks:
                try:
                    await check(self)
                except AssertionError:              
                    error_message = traceback.format_exc().split(
                        '\n'
                    )[-3].strip()

                    self.error = f'Check - {error_message} - for action - {self.name} - failed.'

    @property
    def fields(self):
        return ['name', 'stage', 'time', 'succeeded']

    @property
    def values(self):
        return [self.name, self.stage, self.time, self.success]

    @property
    def record(self):
        return {
            'name': self.name,
            'stage': self.stage,
            'time': self.time,
            'succeeded': self.success
        }

    @property
    def success(self):
        return self.error is None

    def tags_to_dict(self):
        return {
            tag.name: tag.value for tag in self.tags
        }


from.types import ErrorCodes


class StreamError(Exception):
    error_code = None

class StreamResetException(Exception):
    pass


class NoSuchStreamError(Exception):
    """
    A stream-specific action referenced a stream that does not exist.

    .. versionchanged:: 2.0.0
       Became a subclass of :class:`ProtocolError
       <h2.exceptions.ProtocolError>`
    """
    def __init__(self, stream_id):
        #: The stream ID corresponds to the non-existent stream.
        self.stream_id = stream_id


class StreamClosedError(Exception):
    """
    A more specific form of
    :class:`NoSuchStreamError <h2.exceptions.NoSuchStreamError>`. Indicates
    that the stream has since been closed, and that all state relating to that
    stream has been removed.
    """

    __slots__ = (
        'stream_id',
        'error_code',
        'events'
    )

    def __init__(self, stream_id):
        #: The stream ID corresponds to the nonexistent stream.
        self.stream_id = stream_id

        #: The relevant HTTP/2 error code.
        self.error_code = ErrorCodes.STREAM_CLOSED

        # Any events that internal code may need to fire. Not relevant to
        # external users that may receive a StreamClosedError.
        self._events = []



from typing import Callable
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.graphs.hooks.registry.registry_types import InternallHook


class Internal:
    is_internal = True

    def __init__(self) -> None:
        pass

    def __call__(self, call: Callable=None) -> InternallHook:

        hook_fullname = call.__qualname__
        stage_name, hook_shortname = hook_fullname.split('.')

        hook = InternallHook(
            hook_fullname,
            call.__name__,
            call,
            hook_type=HookType.INTERNAL
        )

        registrar.reserved[stage_name][hook_shortname] = hook

        return call


class WorkerPipelineConfig:
    def __init__(self):
        self.executor_config = {}
        self.distributed_config = {}
        self.reporter_config = {}
        self.actions = {}
        self.as_server = False
        self.runner_mode = 'local'
        self.embedded_stats = False
        self.log_level = 'info'



import asyncio
import time
from hedra.tools.data_structures import AsyncList
from typing import List, Dict, Any
from playwright.async_api import async_playwright
from playwright.async_api import Geolocation
from .command import PlaywrightCommand
from .result import PlaywrightResult
from .command_librarian import CommandLibrarian

class ContextGroup:

    __slots__ = (
        'browser_type',
        'device_type',
        'locale',
        'geolocation',
        'permissions',
        'color_scheme',
        'concurrency',
        'librarians',
        'config',
        'options',
        'contexts',
        'sem'
    )

    def __init__(
        self, 
        browser_type: str=None, 
        device_type: str=None, 
        locale: str=None, 
        geolocation: Geolocation=None, 
        permissions: List[str]=None, 
        color_scheme: str=None, 
        concurrency: int=None,
        options: Dict[str, Any]=None
    ) -> None:
        self.browser_type = browser_type
        self.device_type = device_type
        self.locale = locale
        self.geolocation = geolocation
        self.permissions = permissions
        self.color_scheme = color_scheme
        self.concurrency = concurrency
        self.librarians: List[CommandLibrarian] = []
        self.config = {}
        self.options = options
        self.contexts = []
        self.sem = asyncio.Semaphore(value=concurrency)

    async def create(self) -> None:

        playwright = await async_playwright().start()

        if self.browser_type == "safari" or self.browser_type == "webkit":
            self.browser = await playwright.webkit.launch()

        elif self.browser_type == "firefox":
            self.browser = await playwright.firefox.launch()

        else:
            self.browser = await playwright.chromium.launch()


        self.config = {}

        if self.device_type:
            device = playwright.devices[self.device_type]

            self.config = {
                **device,
                **self.options
            }

        if self.locale:
            self.config['locale'] = self.locale

        if self.geolocation:
            self.config['geolocation']: Geolocation = self.geolocation

        if self.permissions:
            self.config['permissions'] = self.permissions

        if self.color_scheme:
            self.config['color_scheme'] = self.color_scheme


        has_options = len(self.config) > 0

        for _ in range(self.concurrency):

            if has_options:        
                context = await self.browser.new_context(
                    **self.config
                )
    
            else:
                context = await self.browser.new_context()
                
            self.contexts.append(context)
            page = await context.new_page()
            command_librarian = CommandLibrarian(page)

            self.librarians.append(command_librarian)

    async def execute(self, command: PlaywrightCommand):
        result = PlaywrightResult(command)
        await self.sem.acquire()
        start = 0
        librarian = self.librarians.pop()

        try:
            playwright_command = librarian.get(command.command)

            start = time.time()

            result.data = await playwright_command(command)

            elapsed = time.time() - start

            result.time = elapsed

            self.librarians.append(librarian)
            self.sem.release()

            return result

        except Exception as e:
            elapsed = time.time() - start
            result.time = elapsed
            result.error = e

            self.librarians.append(librarian)
            self.sem.release()

            return result

    async def execute_batch(self, command: PlaywrightCommand, timeout: float=None):
        return await asyncio.wait([
            self.execute(command) async for _ in AsyncList(range(self.concurrency))
        ], timeout=timeout)

    async def close(self):
        try:
            for context in self.contexts:
                await context.close()
                
            await self.browser.close()
        
        except Exception:
            pass



from types import SimpleNamespace
from typing import Dict
from pydantic import BaseModel

from hedra.reporting.types.common.types import ReporterTypes


try:
    import sqlalchemy
except Exception:
    sqlalchemy = SimpleNamespace(Column=None)


class MySQLConfig(BaseModel):
    host: str='127.0.0.1'
    database: str
    username: str
    password: str
    events_table: str='events'
    metrics_table: str='metrics'
    custom_fields: Dict[str, sqlalchemy.Column]={}
    reporter_type: ReporterTypes=ReporterTypes.MySQL

    class Config:
        arbitrary_types_allowed = True


import uuid
import asyncio
from typing import Generic, TypeVar
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types import (
    MercuryGraphQLClient,
    MercuryGraphQLHTTP2Client,
    MercuryGRPCClient,
    MercuryHTTP2Client,
    MercuryHTTPClient,
    MercuryPlaywrightClient,
    MercuryTaskRunner,
    MercuryUDPClient,
    MercuryWebsocketClient
)
from hedra.core.engines.types.graphql import GraphQLAction, GraphQLResult
from hedra.core.engines.types.graphql_http2 import GraphQLHTTP2Action, GraphQLHTTP2Result
from hedra.core.engines.types.grpc import GRPCAction, GRPCResult
from hedra.core.engines.types.http import HTTPAction, HTTPResult
from hedra.core.engines.types.http2 import HTTP2Action, HTTP2Result
from hedra.core.engines.types.playwright import PlaywrightCommand, PlaywrightResult
from hedra.core.engines.types.task import Task, TaskResult
from hedra.core.engines.types.udp import UDPAction, UDPResult
from hedra.core.engines.types.websocket import WebsocketAction, WebsocketResult
from hedra.core.engines.client.store import ActionsStore
from hedra.logging import HedraLogger

S = TypeVar(
    'S',
    MercuryGraphQLClient,
    MercuryGraphQLHTTP2Client,
    MercuryGRPCClient,
    MercuryHTTPClient,
    MercuryHTTP2Client,
    MercuryPlaywrightClient,
    MercuryTaskRunner,
    MercuryUDPClient,
    MercuryWebsocketClient
)
A = TypeVar(
    'A',
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    PlaywrightCommand,
    Task,
    UDPAction,
    WebsocketAction
)
R = TypeVar(
    'R',
    GraphQLResult,
    GraphQLHTTP2Result,
    GRPCResult,
    HTTPResult,
    HTTP2Result,
    PlaywrightResult,
    TaskResult,
    UDPResult,
    WebsocketResult
)


class BaseClient(Generic[S, A, R]):
    initialized=False
    setup=False

    def __init__(self) -> None:
        self.initialized = True
        self.metadata_string: str = None
        self.client_id = str(uuid.uuid4())
        self.session: S =  None
        self.request_type :RequestTypes = None
        self.client_type: str = None

        self.actions: ActionsStore = None
        self.next_name = None
        self.intercept = False

        self.logger = HedraLogger()
        self.logger.initialize()

    async def _execute_action(self, action: A) -> R:
        await self.logger.filesystem.aio['hedra.core'].debug(
            f'{self.metadata_string} - {self.client_type} Client {self.client_id} - Preparing Action - {action.name}:{action.action_id}'
        )
        await self.session.prepare(action)

        await self.logger.filesystem.aio['hedra.core'].debug(
            f'{self.metadata_string} - {self.client_type} Client {self.client_id} - Prepared Action - {action.name}:{action.action_id}'
        )

        if self.intercept:
            await self.logger.filesystem.aio['hedra.core'].debug(
                f'{self.metadata_string} - {self.client_type} Client {self.client_id} - Initiating suspense for Action - {action.name}:{action.action_id} - and storing'
            )
            self.actions.store(self.next_name, action, self.session)
            
            loop = asyncio.get_event_loop()
            self.waiter = loop.create_future()
            await self.waiter

        return await self.session.execute_prepared_request(action)


import uuid
from typing import Dict, List, Union
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.engines.client.config import Config
from hedra.core.personas.types.types import PersonaTypes


class SequencedPersona(DefaultPersona):

    def __init__(self, config: Config):
        super(SequencedPersona, self).__init__(config)

        self.persona_id = str(uuid.uuid4())
        self.type = PersonaTypes.SEQUENCE

    def setup(self, hooks: Dict[HookType, List[Union[ActionHook, TaskHook]]], metadata_string: str):

        self.metadata_string = f'{metadata_string} Persona: {self.type.capitalize()}:{self.persona_id} - '
        
        sequence = sorted(
            hooks.get(HookType.ACTION),
            key=lambda action: action.metadata.order
        )

        self.actions_count = len(sequence)
        self._hooks = sequence



import os
import grpc

from broadkast.grpc import BroadkastClient
from broadkast.discovery import Query
from hedra.runners.worker_services.leader_registry import LeaderRegistry
from hedra.runners.utils.connect_timeout import connect_with_retry_async
from hedra.runners.bootstrap_services.proto import (
    BootstrapServerStub
)
from broadkast.grpc.proto import (
    ServiceDiscoveryResponse
)


class BoostrapClient:

    def __init__(self, config) -> None:
        logger = Logger()
        self.session_logger = logger.generate_logger()

        self.use_bootstrap = config.distributed_config.get(
            'use_bootstrap'
        )

        self.bootsrap_ip = config.distributed_config.get(
            'bootstrap_ip',
            os.getenv('BOOSTRAP_SERVER_IP')
        )

        self.bootstrap_port = config.distributed_config.get(
            'bootstrap_port',
            os.getenv('BOOTSTRAP_SERVER_PORT', "8711")
        )

        if self.bootsrap_ip and self.bootstrap_port:
            self.bootstrap_service_address = f'{self.bootsrap_ip}:{self.bootstrap_port}'

        else:
            self.bootstrap_service_address = config.distributed_config.get(
                'bootstrap_address',
                os.getenv(
                    'BOOTSTRAP_SERVER_ADDRESS'
                )
            )

        self.registry = LeaderRegistry([])
        self.channel = grpc.aio.insecure_channel(self.bootstrap_service_address)
        self.service = BootstrapServerStub(self.channel)
        self.broadkast_client = BroadkastClient(config=config.distributed_config)

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def discover(self, host_address: str=None) -> None:
        discovered: ServiceDiscoveryResponse = await self.broadkast_client.discover_service(
            'hedra_leader',
            query=Query(
                labels=[
                    {
                        'name': 'app',
                        'value': 'hedra'
                    },
                    {
                        'name': 'role',
                        'value': 'leader'
                    }
                ]
            ),
            service_port='6669'
        )

        for instance in discovered.instances:
            await self.registry.register_leader(instance.instance_address)

    




import os
import json
from urllib.parse import urlparse
from hedra.projects.management import GraphManager
from hedra.projects.management.graphs.actions import RepoConfig
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def create_project(
    url: str, 
    project_name: str,
    path: str,
    username: str, 
    password: str,
    bypass_connection_validation: bool,
    connection_validation_retries: int,
    log_level: str,
):

    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()

    hedra_config_filepath = os.path.join(
        path,
        '.hedra.json'
    )

    logger['console'].sync.info(f'Checking if project exists at - {path}...')

    project_uninitialized = os.path.exists(hedra_config_filepath) is False
    if project_uninitialized:

        logger['console'].sync.info('No project found! Creating project directories and files.')

        project_directory = os.path.join(path, project_name)
        if os.path.exists(project_directory) is False:
            os.mkdir(project_directory)
        
        package_file = open(f'{project_directory}/__init__.py', 'w')
        package_file.close()

        tests_directory = os.path.join(project_directory, 'tests')
        os.mkdir(tests_directory)

        package_file = open(f'{tests_directory}/__init__.py', 'w')
        package_file.close()

        plugins_directory = os.path.join(project_directory, 'plugins')
        os.mkdir(plugins_directory)

        package_file = open(f'{plugins_directory}/__init__.py', 'w')
        package_file.close()

        log_directory = f'{os.getcwd()}/logs'

        if os.path.exists(log_directory) is False:
            os.mkdir(log_directory)

        elif os.path.isdir(log_directory) is False:
            os.remove(log_directory)
            os.mkdir(log_directory)

        parsed_url = urlparse(url)
        repo_url = f'{parsed_url.scheme}://{username}:{password}@{parsed_url.hostname}{parsed_url.path}'

        logger['console'].sync.info('Initializing project manager.')

        repo_config = RepoConfig(
            path,
            repo_url,
            branch='main',
            remote='origin',
            username=username,
            password=password
        )

        manager = GraphManager(repo_config, log_level=log_level)
        discovered = manager.discover_graph_files()

        new_graphs_count = len(discovered['graphs'])
        new_plugins_count = len(discovered['plugins'])

        logger['console'].sync.info(f'Found - {new_graphs_count} - new graphs and - {new_plugins_count} - plugins.')

        logger['console'].sync.info(f'Linking project to remote at - {url}.')

        workflow_actions = [
            'initialize',
            'create-gitignore'
        ]

        manager.execute_workflow(workflow_actions)

        hedra_config = {
            "name": project_name,
            "core": {
                "bypass_connection_validation": bypass_connection_validation,
                "connection_validation_retries": connection_validation_retries
            },
            'project': {
                'project_url': url,
                'project_username': username,
                'project_password': password
            },
            **discovered
        }

        logger['console'].sync.info('Saving project state to .hedra.json config.')

        with open(hedra_config_filepath, 'w') as hedra_config_file:
                json.dump(hedra_config, hedra_config_file, indent=4)

        logger['console'].sync.info('Project created!')

    else:
        logger['console'].sync.info(f'Found existing project at - {path}. Exiting...\n')



from .graphql_event import GraphQLEvent
from .grpc_event import GRPCEvent
from .http_event import HTTPEvent
from .http2_event import HTTP2Event
from .playwright_event import PlaywrightEvent
from .task_event import TaskEvent
from .udp_event import UDPEvent
from .websocket_event import WebsocketEvent


from hedra.core.engines.types.http2.streams.changed_setting import ChangedSetting
from hedra.core.engines.types.http2.streams.stream_settings import SettingCodes
from .base_event import BaseEvent


class RemoteSettingsChanged(BaseEvent):
    event_type='REMOTE_SETTINGS_CHANGED'
    __slots__ = (
        'changed_settings'
    )
    """
    The RemoteSettingsChanged event is fired whenever the remote peer changes
    its settings. It contains a complete inventory of changed settings,
    including their previous values.

    In HTTP/2, settings changes need to be acknowledged. h2 automatically
    acknowledges settings changes for efficiency. However, it is possible that
    the caller may not be happy with the changed setting.

    When this event is received, the caller should confirm that the new
    settings are acceptable. If they are not acceptable, the user should close
    the connection with the error code :data:`PROTOCOL_ERROR
    <h2.errors.ErrorCodes.PROTOCOL_ERROR>`.

    .. versionchanged:: 2.0.0
       Prior to this version the user needed to acknowledge settings changes.
       This is no longer the case: h2 now automatically acknowledges
       them.
    """
    def __init__(self):
        #: A dictionary of setting byte to
        #: :class:`ChangedSetting <h2.settings.ChangedSetting>`, representing
        #: the changed settings.
        self.changed_settings = {}

    @classmethod
    def from_settings(cls, old_settings, new_settings):
        """
        Build a RemoteSettingsChanged event from a set of changed settings.

        :param old_settings: A complete collection of old settings, in the form
                             of a dictionary of ``{setting: value}``.
        :param new_settings: All the changed settings and their new values, in
                             the form of a dictionary of ``{setting: value}``.
        """
        e = cls()
        for setting, new_value in new_settings.items():

            try:
                setting = SettingCodes(setting)
            except ValueError:
                pass

            original_value = old_settings.get(setting)
            change = ChangedSetting(setting, original_value, new_value)
            e.changed_settings[setting] = change

        return e

    def __repr__(self):
        return "<RemoteSettingsChanged changed_settings:{%s}>" % (
            ", ".join(repr(cs) for cs in self.changed_settings.values()),
        )


import asyncio
import traceback
import h2.settings
from ssl import SSLContext
from typing import Tuple, Optional, Union
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.protocols.tcp import TCPConnection
from .stream import Stream
from .frames import FrameBuffer
from .frames.types.base_frame import Frame

class HTTP2Connection:

    __slots__ = (
        'timeouts',
        'connected',
        'init_id',
        'reset_connection',
        'stream_type',
        'init_id',
        'stream_id',
        'concurrency',
        'dns_address',
        'port',
        'connection',
        'lock',
        'stream',
        'local_settings',
        'remote_settings',
        'outbound_flow_control_window',
        'local_settings_dict',
        'remote_settings_dict',
        'settings_frame',
        'headers_frame',
        'window_update_frame'
    )

    def __init__(self, stream_id: int, timeouts: Timeouts, concurrency: int, reset_connection: bool, stream_type: RequestTypes) -> None:
        self.timeouts = timeouts
        self.connected = False
        self.init_id = stream_id
        self.reset_connection = reset_connection
        self.stream_type = stream_type

        if self.init_id%2 == 0:
            self.init_id += 1

        self.stream_id = 0
        self.concurrency = concurrency
        self.dns_address = None
        self.port = None

        self.connection = TCPConnection(stream_type)
        self.lock = asyncio.Lock()
        self.stream = Stream(self.stream_id, self.timeouts)

        self.local_settings = h2.settings.Settings(
            client=True,
            initial_values={
                h2.settings.SettingCodes.ENABLE_PUSH: 0,
                h2.settings.SettingCodes.MAX_CONCURRENT_STREAMS: concurrency,
                h2.settings.SettingCodes.MAX_HEADER_LIST_SIZE: 65535,
            }
        )
        self.remote_settings = h2.settings.Settings(
            client=False
        )

        self.outbound_flow_control_window = self.remote_settings.initial_window_size

        del self.local_settings[h2.settings.SettingCodes.ENABLE_CONNECT_PROTOCOL]

        self.local_settings_dict = {setting_name: setting_value for setting_name, setting_value in self.local_settings.items()}
        self.remote_settings_dict = {setting_name: setting_value for setting_name, setting_value in self.remote_settings.items()}

        self.settings_frame = Frame(0, 0x04, settings=self.local_settings_dict)
        self.headers_frame = Frame(self.init_id, 0x01)
        self.headers_frame.flags.add('END_HEADERS')
        
        self.window_update_frame = Frame(self.init_id, 0x08, window_increment=65536)

        self.stream.connection_data.extend(self.settings_frame.serialize())

    async def connect(self, 
        hostname: str, 
        dns_address: str,
        port: int, 
        socket_config: Tuple[int, int, int, int, Tuple[int, int]], 
        ssl: Optional[SSLContext]=None,
        timeout: Optional[float] = None
    ) -> Union[Stream, Exception]:
        
            try:
                if self.connected is False or self.dns_address != dns_address or self.reset_connection:
                    reader, writer = await asyncio.wait_for(
                        self.connection.create_http2(
                            hostname,
                            socket_config=socket_config,
                            ssl=ssl
                        ),
                        timeout=timeout
                    )

                    self.connected = True
                    self.stream_id = self.init_id
                    self.dns_address = dns_address
                    self.port = port

                    self.stream.reader = reader
                    self.stream.writer = writer
                  
                    self.stream.headers_frame = self.headers_frame
                    self.stream.window_frame = self.window_update_frame

                else:

                    self.stream_id += 2# self.concurrency
                    if self.stream_id%2 == 0:
                        self.stream_id += 1

                    self.stream.stream_id = self.stream_id
                    self.stream.headers_frame.stream_id = self.stream_id
                    self.stream.window_frame.stream_id = self.stream_id

                self.stream.frame_buffer = FrameBuffer()
                return self.stream
            
            except asyncio.TimeoutError:
                raise Exception('Connection timed out.')

            except ConnectionResetError:
                raise Exception('Connection reset.')

            except Exception as e:
                raise e

    async def close(self):
        await self.connection.close()


import asyncio
from typing import Coroutine, List, Generic, TypeVar


A = TypeVar('A')


class Hooks(Generic[A]):

    __slots__ = (
        'before',
        'after',
        'checks',
        'notify',
        'listen',
        'listeners',
        'channels'
    )

    def __init__(self) -> None:
        self.before: Coroutine = None
        self.after: Coroutine = None
        self.checks: List[Coroutine] = []
        self.notify = False
        self.listen = False
        self.channel_events: List[asyncio.Event] = []
        self.listeners: List[A] = []
        self.channels: List[Coroutine] = []

    def to_names(self):

        names = {}

        if self.before:
            names['before'] = self.before.__qualname__

        if self.after:
            names['after'] = self.after.__qualname__

        check_names = []
        for check in self.checks:
            check_names.append(check.__qualname__)

        names['checks'] = check_names

        return names

    def to_serializable(self):
        return {
            'notify': self.notify,
            'listen': self.listen,
            'listeners': [
                listener.name for listener in self.listeners
            ],
            'names': self.to_names()
        }


import asyncio
import uuid
import functools
from typing import List
from numpy import float32, float64, int16, int32, int64
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet

try:
    from prometheus_client.exposition import basic_auth_handler
    from prometheus_client import (
        CollectorRegistry,
        push_to_gateway,
    )
    from .prometheus_metric import PrometheusMetric
    from prometheus_client.core import REGISTRY
    from .prometheus_config import PrometheusConfig
    has_connector = True

except Exception:
    PrometheusConfig = None
    has_connector = False
    basic_auth_handler = lambda: None


class Prometheus:

    def __init__(self, config: PrometheusConfig) -> None:
        self.pushgateway_address = config.pushgateway_address
        self.auth_request_method = config.auth_request_method
        self.auth_request_timeout = config.auth_request_timeout
        self.auth_request_data = config.auth_request_data
        self.username = config.username
        self.password = config.password
        self.namespace = config.namespace
        self.job_name = config.job_name

        self.registry = None
        self._auth_handler = None
        self._has_auth = False
        self._loop = asyncio.get_event_loop()
        
        self.types_map = {
            'total': 'count',
            'succeeded': 'count',
            'failed': 'count',
            'actions_per_second': 'gauge',
            'median': 'gauge',
            'mean': 'gauge',
            'variance': 'gauge',
            'stdev': 'gauge',
            'minimum': 'gauge',
            'maximum': 'gauge'
        }

        self._events = {}
        self._metrics = {}
        self._shared_metrics = {}
        self._custom_metrics = {}
        self._errors = {}

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        
    async def connect(self) -> None:

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Prometheus Pushgateway at: {self.pushgateway_address}')

        self.registry = CollectorRegistry()
        REGISTRY.register(self.registry)

        if self.username and self.password:
            self._has_auth = True

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Prometheus Pushgateway at: {self.pushgateway_address}')
            

    def _generate_auth(self) -> basic_auth_handler:
        return basic_auth_handler(
            self.pushgateway_address,
            self.auth_request_method,
            self.auth_request_timeout,
            {
                'Content-Type': 'application/json'
            },
            self.auth_request_data,
            username=self.username,
            password=self.password
        )
    
    async def _submit_to_pushgateway(self):
        if self._has_auth:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushing to secure Prometheus Pushgateway via HTTPS')
            await self._loop.run_in_executor(
                None,
                functools.partial(
                    push_to_gateway,
                    self.pushgateway_address,
                    job=self.job_name,
                    registry=self.registry,
                    handler=self._generate_auth
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushed to secure Prometheus Pushgateway via HTTPS')

        else:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushing to Prometheus Pushgateway via HTTP')
            await self._loop.run_in_executor(
                None,
                functools.partial(
                    push_to_gateway,
                    self.pushgateway_address,
                    job=self.job_name,
                    registry=self.registry
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Pushed to Prometheus Pushgateway via HTTP')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Prometheus - Namespace: {self.namespace}')

        for event in events:
            
            record = {
                f'{event.name}_time': event.time,
                f'{event.name}_success': 1 if event.success else 0,
                f'{event.name}_failed': 1 if event.success is False else 0
            }

            if self._events.get(event.name) is None:

                fields = {}

                for event_field in record.keys():
                    metric_type = self.types_map.get(event_field)
                
                    metric = PrometheusMetric(
                        f'{event.name}_{event_field}',
                        metric_type,
                        metric_description=f'{event.name} {event_field}',
                        metric_labels=[],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )

                    metric.create_metric()

                    fields[event_field] = metric

                self._events[event.name] = fields
                
            for event_field, event_value in record.items():
                if event_value and event_field in self.types_map:
                    self._events[event.name][event_field].update(event_value)

            await self._submit_to_pushgateway()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Prometheus - Namespace: {self.namespace}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Prometheus - Namespace: {self.namespace}')
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]
            
            shared_metrics = self._shared_metrics.get(metrics_set.name)
            if shared_metrics is None:
                shared_metrics = {}

                for field in metrics_set.common_stats.keys():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')

                    metric_name = f'{metrics_set.name}_{field}'.replace('.', '_')
                    metric_type = self.types_map.get(field)

                    prometheus_metric = PrometheusMetric(
                        metric_name,
                        metric_type,
                        metric_description=f'{metrics_set.name} {field}',
                        metric_labels=[
                            *tags,
                            f'stage:{metrics_set.stage}',
                            'group:common'
                        ],
                        metric_namespace=self.namespace,
                        registry=self.registry
                    )
                    prometheus_metric.create_metric()
                    
                    shared_metrics[field] = prometheus_metric
                
                self._shared_metrics[metrics_set.name] = shared_metrics

            self._shared_metrics[metrics_set.name] = shared_metrics
            for field, value in metrics_set.common_stats.items():
                metric = shared_metrics.get(field)
                metric.update(value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Prometheus - Namespace: {self.namespace}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Prometheus - Namespace: {self.namespace}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            tags = [
                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
            ]

            stage_metrics = self._metrics.get(metrics_set.name)
            if stage_metrics is None:
                stage_metrics = {}

                for group_name, group in metrics_set.groups.items():
                    group_metrics = stage_metrics.get(group_name)

                    if group_metrics is None:
                        group_metrics = {}

                        for quantile_name in metrics_set.quantiles:
                            self.types_map[quantile_name] = 'gauge'

                        fields = {}

                        for metric_field in metrics_set.stats_fields:
                            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{metric_field}')
                            metric_type = self.types_map.get(metric_field)
                            metric_name = f'{metrics_set.name}_{group_name}_{metric_field}'.replace('.', '_')

                            prometheus_metric = PrometheusMetric(
                                metric_name,
                                metric_type,
                                metric_description=f'{metrics_set.name} {metric_field}',
                                metric_labels=[
                                    *tags,
                                    f'stage:{metrics_set.stage}',
                                    f'group:{group_name}'
                                ],
                                metric_namespace=self.namespace,
                                registry=self.registry
                            )
                            prometheus_metric.create_metric()

                            fields[metric_field] = prometheus_metric

                        group_metrics[group_name] = fields

                stage_metrics[metrics_set.name] = group_metrics

            self._metrics[metrics_set.name] = stage_metrics
            for group_name, group in metrics_set.groups.items():
                group_metrics = stage_metrics.get(group_name)
                record = group.record

                for field in group_metrics[group_name]:
                    metric_value = record.get(field)
                    field_metric = group_metrics.get(field)
                    field_metric.update(metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Prometheus - Namespace: {self.namespace}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Prometheus - Namespace: {self.namespace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            custom_metrics = self._custom_metrics.get(metrics_set.name)
            if custom_metrics is None:
                custom_metrics = {}

                for custom_group_name, group in metrics_set.custom_metrics.items():
                    group_metrics = custom_metrics.get(custom_group_name)

                    if group_metrics is None:

                        for quantile_name in metrics_set.quantiles.keys():
                            self.types_map[quantile_name] = 'gauge'

                        fields = {}

                        for field, value in group.items():
                            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metric - {metrics_set.name}:{group_name}:{field}')
                            
                            metric_type = None
                            if isinstance(value, (int, int16, int32, int64)):
                                metric_type = 'count'

                            elif isinstance(value, (float, float32, float64)):
                                metric_type = 'gauge'

                            metric_name = f'{metrics_set.name}_{field}'.replace('.', '_')

                            tags = [
                                f'{tag.name}:{tag.value}' for tag in metrics_set.tags
                            ]

                            prometheus_metric = PrometheusMetric(
                                metric_name,
                                metric_type,
                                metric_description=f'{metrics_set.name} {field}',
                                metric_labels=[
                                    *tags,
                                    f'stage:{metrics_set.stage}',
                                    f'group:{custom_group_name}'
                                ],
                                metric_namespace=self.namespace,
                                registry=self.registry
                            )
                            prometheus_metric.create_metric()

                            fields[field] = prometheus_metric

                        group_metrics[custom_group_name] = fields

                custom_metrics[metrics_set.name] = group_metrics

            self._custom_metrics[metrics_set.name] = custom_metrics
            for group_name, group in metrics_set.custom_metrics.items():
                group_metrics = custom_metrics.get(group_name)

                for field in group_metrics[group_name]:
                    metric_value = group.get(field)
                    field_metric = group_metrics.get(field)
                    field_metric.update(metric_value)

        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Prometheus - Namespace: {self.namespace}')
                    
    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Prometheus - Namespace: {self.namespace}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._errors.get(metrics_set.name) is None:
                errors_metric = PrometheusMetric(
                    f'{metrics_set.name}_errors',
                    'count',
                    metric_description=f'Errors for action - {metrics_set.name}.',
                    metric_labels=[],
                    metric_namespace=self.namespace,
                    registry=self.registry
                )

                errors_metric.create_metric()
                self._errors[metrics_set.name] = errors_metric

            for error in metrics_set.errors:              
                errors_metric.update(
                    error.get('count'),
                    labels={
                        'message': error.get('message')
                    }
                )
                
        await self._submit_to_pushgateway()
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Prometheus - Namespace: {self.namespace}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
                


from .base_event import BaseEvent

class StreamReset(BaseEvent):
    event_type='STREAM_RESET'
    __slots__ = ('stream_id', 'error_code', 'remote_reset')
    """
    The StreamReset event is fired in two situations. The first is when the
    remote party forcefully resets the stream. The second is when the remote
    party has made a protocol error which only affects a single stream. In this
    case, h2 will terminate the stream early and return this event.

    .. versionchanged:: 2.0.0
       This event is now fired when h2 automatically resets a stream.
    """
    def __init__(self):
        #: The Stream ID of the stream that was reset.
        self.stream_id = None

        #: The error code given. Either one of :class:`ErrorCodes
        #: <h2.errors.ErrorCodes>` or ``int``
        self.error_code = None

        #: Whether the remote peer sent a RST_STREAM or we did.
        self.remote_reset = True

    def __repr__(self):
        return "<StreamReset stream_id:%s, error_code:%s, remote_reset:%s>" % (
            self.stream_id, self.error_code, self.remote_reset
        )



from typing import Awaitable
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.graphs.stages import Validate
from hedra.core.graphs.hooks import validate


class ValidateStage(Validate):

    @validate('<action_name_here>')
    async def validate_action(self, action_or_task: Awaitable[BaseResult]):
        result: BaseResult = await action_or_task()
        assert result is not None
        assert result.action_id is not None
        assert result.error is False


from typing import Any, List, Type, Callable, Awaitable
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook


class AfterHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: List[str]
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.AFTER
        )

        self.call: Type[self._call] = self._call
        self.names = list(set(names))





import binascii
from .base_event import BaseEvent


class DataReceived(BaseEvent):
    event_type='DATA_RECEIVED'

    __slots__ = (
        'stream_id',
        'data',
        'flow_controlled_length',
        'stream_ended'
    )

    def __init__(self):
        #: The Stream ID for the stream this data was received on.
        self.stream_id = None

        #: The data itself.
        self.data = None

        #: The amount of data received that counts against the flow control
        #: window. Note that padding counts against the flow control window, so
        #: when adjusting flow control you should always use this field rather
        #: than ``len(data)``.
        self.flow_controlled_length = None

        #: If this data chunk also completed the stream, the associated
        #: :class:`StreamEnded <h2.events.StreamEnded>` event will be available
        #: here.
        #:
        #: .. versionadded:: 2.4.0
        self.stream_ended = None

    def __repr__(self):
        
        decoded_data = None
        if self.data:
            decoded_data = binascii.hexlify(self.data[:20]).decode('ascii')

        return (
            "<DataReceived stream_id:%s, "
            "flow_controlled_length:%s, "
            "data:%s>" % (
                self.stream_id,
                self.flow_controlled_length,
                decoded_data,
            )
        )


import asyncio
from typing import Dict, List
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.setup import Setup
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.logging import HedraLogger


async def setup_transition(current_stage: Setup, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

    execute_stages = current_stage.context.stages.get(StageTypes.EXECUTE).items()
    paths = current_stage.context.paths.get(current_stage.name)
    path_lengths: Dict[str, int] = current_stage.context.path_lengths.get(current_stage.name)
    visited = current_stage.context.visited

    execute_stages: Dict[str, Execute] = {
        stage_name: stage for stage_name, stage in execute_stages if stage_name in paths and stage_name not in visited
    }

    setup_candidates = {}
    
    setup_stages: Dict[str, Setup] = current_stage.context.stages.get(StageTypes.SETUP)
    execute_stages: Dict[str, Execute] = current_stage.context.stages.get(StageTypes.EXECUTE)
    following_setup_stage_distances = [
        path_length for stage_name, path_length in path_lengths.items() if stage_name in setup_stages
    ]

    for stage_name in path_lengths.keys():
        stage_distance = path_lengths.get(stage_name)

        if stage_name in execute_stages:

            if len(following_setup_stage_distances) > 0 and stage_distance < min(following_setup_stage_distances):
                setup_candidates[stage_name] = execute_stages.get(stage_name)

            elif len(following_setup_stage_distances) == 0:
                setup_candidates[stage_name] = execute_stages.get(stage_name)

    current_stage.generation_setup_candidates = len(setup_candidates)
    stages = {}
    next_stage_decendants = current_stage.context.paths.get(next_stage.name)
    path_decendants = [
        next_stage.name,
        *next_stage_decendants
    ]

    if len(current_stage.apply_to_stages) > 0:
        for stage_name in current_stage.apply_to_stages:
            execute_stage: Execute = execute_stages.get(stage_name)
            if execute_stage.state == StageStates.INITIALIZED and execute_stage.name in path_decendants:
                execute_stage.state = StageStates.SETTING_UP
                stages[stage_name] = execute_stage

    else:

        generation_setup_candidates = [
            stage_name for stage_name in current_stage.generation_stage_names if stage_name in current_stage.context.stages.get(StageTypes.SETUP)
        ]

        user_specified_setup_candidates = []
        all_setup_stages: Dict[str, Setup] = current_stage.context.stages.get(StageTypes.SETUP)
        for stage_name in generation_setup_candidates:
            stage: Setup = all_setup_stages.get(stage_name)

            user_specified_setup_candidates.extend(stage.apply_to_stages)

        for execute_stage_name, execute_stage in setup_candidates.items():
            if execute_stage.state == StageStates.INITIALIZED and execute_stage.name in path_decendants and execute_stage.name not in user_specified_setup_candidates:
                execute_stage.state = StageStates.SETTING_UP
                stages[execute_stage_name] = execute_stage

    current_stage.stages = stages

    if current_stage.timeout:

        setup_results = await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)

    else:

        setup_results = await current_stage.run()

    for execute_stage in stages.values():
        execute_stage.state = StageStates.SETUP

    current_stage.context.stages[StageTypes.EXECUTE].update(setup_results)

    await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
    
    next_stage.context = current_stage.context
    current_stage = None


async def setup_to_validate_transition(current_stage: Stage, next_stage: Stage):
    
    try:

        await setup_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.VALIDATE


async def setup_to_optimize_transition(current_stage: Stage, next_stage: Stage):

    try:

        await setup_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.OPTIMIZE


async def setup_to_execute_transition(current_stage: Stage, next_stage: Stage):

    try:

        await setup_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.EXECUTE


async def setup_to_checkpoint_transition(current_stage: Stage, next_stage: Stage):

    try:

        await setup_transition(current_stage, next_stage)

        next_stage.data = {
            'stages': current_stage.context.stages[StageTypes.EXECUTE],
            'reporter_config': current_stage.context.reporting_config
        }
        
        next_stage.previous_stage = current_stage.name
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR
    
    return None, StageTypes.CHECKPOINT


async def setup_to_wait_transition(current_stage: Stage, next_stage: Stage):

    try:

        await setup_transition(current_stage, next_stage)
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR
    
    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR

    return None, StageTypes.WAIT


from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    TelegrafStatsDConfig
)


class SubmitTelegrafStatsDResultsStage(Submit):
    config=TelegrafStatsDConfig(
        host='0.0.0.0',
        port=8125
    )


from hedra.core.graphs.stages.base.stage import Stage


class InvalidTransitionError(Exception):
    
    def __init__(self, from_stage: Stage, to_stage: Stage) -> None:

        from_stage_name = from_stage.__class__.__name__
        to_stage_name = to_stage.__class__.__name__
        from_stage_type = from_stage.stage_type.name.capitalize()
        to_stage_type = to_stage.stage_type.name.capitalize()

        super().__init__(
            f'Error - {from_stage_type} stage {from_stage_name} cannot preceed {to_stage_type} stage {to_stage_name}.'
        )

class IdleTranstionError(Exception):
    def __init__(self, stage: Stage) -> None:

        stage_name = stage.__class__.__name__
        stage_type = stage.stage_type.name.capitalize()

        super().__init__(
            f"Error - {stage_type} stage {stage_name} cannot be the initial stage. Make sure you have a Setup class and set it as {stage_name}'s dependency!"
        )

class IsolatedStageError(Exception):

    def __init__(self, stage: Stage) -> None:

        stage_name = stage.__class__.__name__
        stage_type = stage.stage_type.name.capitalize()

        super().__init__(
            f'Error - {stage_type} stage {stage_name} has no dependencies and is thus unreachable by the execution graph.'
        )


class StageTimeoutError(Exception):

    def __init__(self, from_stage: Stage) -> None:
        self.from_stage = from_stage

        super().__init__(
            f'Stage Timeout Exception - Stage {from_stage.name} of type {from_stage.stage_type.name}:\nStage exceeded provided timeout of {from_stage.timeout} seconds.'
        )


class StageExecutionError(Exception):

    def __init__(self, from_stage: Stage, to_stage: Stage, message: str) -> None:

        self.from_stage = from_stage
        self.to_stage = to_stage

        super().__init__(
            f'Stage Execution Exception - Stage {from_stage.name} of type {from_stage.stage_type.name}:\n Encountered error - {message} - while transitioning to {to_stage.stage_type.name} stage - {self.to_stage.name}.'
        )


from typing import Dict, Any, Union
from hedra.core.engines import registered_engines
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.task import Task, MercuryTaskRunner
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.engines.types.common.types import RequestTypes
from hedra.logging import HedraLogger


async def assemble_task(
    hook: Union[ActionHook, TaskHook],
    hook_action: Dict[str, Any],
    persona: DefaultPersona,
    config: Config,
    metadata_string: str
):

    logger = HedraLogger()
    logger.initialize()

    hook_type_name = RequestTypes.TASK.capitalize()
    task_name = hook_action.get('name')
    
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Assembling {hook_type_name} Hook - {task_name}')
    
    task_hooks = hook_action.get('hooks', {})
    task_hook_names = task_hooks.get('names')

    before_hook_name = task_hook_names.get('before')
    after_hook_name = task_hook_names.get('after')
    check_hook_names = task_hook_names.get('checks', [])
    listener_names = task_hook_names.get('listeners', [])
    channel_hook_names = task_hook_names.get('channels', [])

    if before_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {task_name} - Found Before Hook - {before_hook_name}'
        )

    if after_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {task_name} - Found After Hook - {after_hook_name}'
        )

    for check_hook_name in check_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {task_name} - Found Check Hook - {check_hook_name}'
        )

    for channel_hook_name in channel_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {task_name} - Found Channel Hook - {channel_hook_name}'
        )

    for listener_name in listener_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {task_name} - Found Listener - {listener_name}'
        )

    hook.session: MercuryTaskRunner = registered_engines.get(RequestTypes.TASK)(
        concurrency=persona.batch.size,
        timeouts=hook_action.get('timeouts')
    )

    task_hook = registrar.all.get(task_name)

    hook.action: Task = Task(
        task_name,
        task_hook.call,
        source=hook.config.env,
        user=hook.config.user,
        tags=hook.config.tags
    )


    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {task_name} - Env: {hook.action.source}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {task_name} - User: {hook.action.metadata.user}')

    for tag in hook.action.metadata.tags_to_string_list():
        await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {task_name} - Tag: {tag}')
    
    hook.action.hooks.notify = task_hooks.get('notify', False)
    hook.action.hooks.listen = task_hooks.get('listen', False)


    if before_hook_name:
        before_hook = registrar.all.get(before_hook_name)
        hook.action.hooks.before = before_hook.call

    if after_hook_name:
        after_hook = registrar.all.get(after_hook_name)
        hook.action.hooks.after = after_hook.call

    hook.action.hooks.checks = []
    hook.action.hooks.checks.extend(check_hook_names)

    hook.action.hooks.channels = []
    for channel_hook_name in channel_hook_names:
        channel = registrar.all.get(channel_hook_name)
        hook.action.hooks.channels.append(channel.call) 

    hook.action.hooks.listeners = listener_names

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {task_name} - Setup complete')

    return hook


import asyncio
import uuid
from .leader_server import LeaderServer


class LeaderManager:

    def __init__(self, config, reporter_config) -> None:
        self.config = config
        self.reporter_config = reporter_config

        if self.leader_type == 'ephemeral-leader':
            self.job_config = config


        self.leader_id = None

        self._live_progress_updates = config.distributed_config.get('live_progress_updates', False)
        
        self.session_logger.info(f'Initializing leader...')


        self.server = LeaderServer(config)
        self.running = True

    async def register(self):

        self.session_logger.info(f'Starting leader at - {self.server.service.job_registry.electorate.leader_ip}:{self.server.service.job_registry.electorate.leader_port}.')

        await self.server.create_server(self.config)
        self.leader_id = await self.server.service.job_registry.register_leaders(self.config)
        if self.leader_id is None:
            self.leader_id = uuid.uuid4()
        
        await self.server.setup(self.leader_id)

        self.session_logger.info(f'\nWaiting for - {self.server.service.job_registry.electorate.registered_services.expected_leaders_count} - leaders.')
        await self.server.service.job_registry.wait_for_leaders()
        await self.server.run()
        
        self.session_logger.info(f'Waiting for - {self.server.service.job_registry.workers.registry.initial_count} - workers.\n')

        await self.server.leader_register_workers()         
        await self.server.service.job_registry.wait_for_workers()

        registered_count = self.server.service.job_registry.workers.registry.count
        expected_count = self.server.service.job_registry.workers.registry.initial_count

        if registered_count == expected_count:
            self.session_logger.info(f'Starting monitor...')
            if self.server.service.job_registry.monitoring is False:
                await self.server.start_monitor()

            self.session_logger.info(f'Startup complete!')

        else:
            raise Exception(f'Error: Timed out waiting for {expected_count} workers.')

    async def wait(self):

        if self.leader_type == 'ephemeral-leader':
            leader_job_id = 'ephemeral-job'

            self.session_logger.info(f'\nLeader - {self.leader_id} - starting ephemeral job.')

            await self.server.service.job_registry.create_ephemeral_job(leader_job_id, self.job_config)
            await self.server.service.job_registry.start_job(leader_job_id)

        self.server.service.job_registry.ready = True

        while self.running:
            new_worker_queue = self.server.service.new_worker_queue
            not_empty = await new_worker_queue.not_empty()

            if not_empty:
                await self.server.add_new_workers(new_worker_queue)
                await new_worker_queue.clear()
                self.server.service.new_worker_queue = new_worker_queue

            if self.leader_type == 'ephemeral-leader':
                job = self.server.service.job_registry[leader_job_id]

                if job.results_task and job.results_task.done():
                    await self.server.service.job_registry.workers.registry.clear_registry()
                    self.running = False

                    self.session_logger.info('Ephemeral job complete!\n')

                    return
            
            await self.server.service.job_registry.workers.check_worker_heartbeat()
            await self.server.service.job_registry.check_for_completed()
            await self.server.service.job_registry.update_leader_active_jobs()

            await asyncio.sleep(1)

    async def stop(self):
        await self.server.stop_server()




import asyncio
import uuid
import functools
import json
import psutil
from datetime import datetime
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet
from concurrent.futures import ThreadPoolExecutor

try:
    import boto3
    from .s3_config import S3Config
    has_connector = True

except Exception:
    boto3 = None
    S3Config = None
    has_connector = False


class S3:

    def __init__(self, config: S3Config) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name
        self.buckets_namespace = config.buckets_namespace
        self.events_bucket_name = config.events_bucket
        self.metrics_bucket_name = config.metrics_bucket
        self.shared_metrics_bucket_name = 'stage-metrics'
        self.errors_bucket_name = 'stage-errors'

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.client = None
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to AWS S3 - Region: {self.region_name}')
        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                's3',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to AWS S3 - Region: {self.region_name}')

    async def submit_events(self, events: List[BaseEvent]):
        
        events_bucket_name = f'{self.buckets_namespace}-{self.events_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Bucket - {events_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events Bucket - {events_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=events_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Events Bucket - {events_bucket_name} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Events Bucket - {events_bucket_name}')


        for event in events:
            timestamp = int(datetime.now().timestamp())
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=events_bucket_name,
                    Key=f'{event.name}-{timestamp}',
                    Body=json.dumps(event.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Bucket - {events_bucket_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        shared_metrics_bucket_name = f'{self.buckets_namespace}-{self.shared_metrics_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Bucket - {shared_metrics_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics Bucket - {shared_metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=shared_metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Shared Metrics Bucket - {shared_metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Shared Metrics Bucket - {shared_metrics_bucket_name}')
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            timestamp = int(datetime.now().timestamp())
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=shared_metrics_bucket_name,
                    Key=f'{metrics_set.name}-{timestamp}',
                    Body=json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        **metrics_set.common_stats
                    })
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Bucket - {shared_metrics_bucket_name}')
    
    async def submit_metrics(self, metrics: List[MetricsSet]):

        metrics_bucket_name = f'{self.buckets_namespace}-{self.metrics_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Bucket - {metrics_bucket_name}')
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics Bucket - {metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Metrics Bucket - {metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Metrics Bucket - {metrics_bucket_name}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                timestamp = int(datetime.now().timestamp())
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.put_object,
                        Bucket=metrics_bucket_name,
                        Key=f'{metrics_set.name}-{timestamp}',
                        Body=json.dumps({
                            **group.record,
                            'group': group_name
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Bucket - {metrics_bucket_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            for custom_group_name, group in metrics_set.custom_metrics.items():

                custom_bucket_name = f'{self.buckets_namespace}-{custom_group_name}-metrics'
                
                try:
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics Bucket - {custom_bucket_name} - if not exists')
                    await self._loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.create_bucket,
                            Bucket=custom_bucket_name,
                            CreateBucketConfiguration={
                                'LocationConstraint': self.region_name
                            }
                        )
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Custom Metrics Bucket - {custom_bucket_name}')

                except Exception:
                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Custom Metrics Bucket - {custom_bucket_name}')

                timestamp = int(datetime.now().timestamp())
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.put_object,
                        Bucket=custom_bucket_name,
                        Key=f'{metrics_set.name}-{custom_group_name}-{timestamp}',
                        Body=json.dumps({
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Bucket - {custom_bucket_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        errors_bucket_name = f'{self.buckets_namespace}-{self.errors_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Bucket - {errors_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Error Metrics Bucket - {errors_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=errors_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Error Metrics Bucket - {errors_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Error Metrics Bucket - {errors_bucket_name}')

        for metrics_set in metrics_sets:
            for error in metrics_set.errors:
                timestamp = int(datetime.now().timestamp())
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.put_object,
                        Bucket=errors_bucket_name,
                        Key=f'{metrics_set.name}-{timestamp}',
                        Body=json.dumps({
                            'metrics_name': metrics_set.name,
                            'metrics_stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Bucket - {errors_bucket_name}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')



import datetime
from typing import Dict, Type, Union
from aiologger.levels import LogLevel
from .logger_types.handers.async_file_handler import RolloverInterval
from .logger_types import (
    Logger, 
    LoggerTypes, 
    LoggerTypesMap,
    AsyncLogger,
    AsyncFilesystemLogger,
    AsyncSpinner,
    SyncLogger,
    SyncFilesystemLogger
)
from .config import LoggingConfig
from .logging_manager import logging_manager


class HedraLogger:

    def __init__(self) -> None:
        self.loggers: Dict[str, Logger[Union[AsyncLogger, AsyncFilesystemLogger, AsyncSpinner], Union[SyncLogger, SyncFilesystemLogger]]] = {}

        self.logger_names = logging_manager.logger_types.names
        self.logger_types = logging_manager.logger_types.types
        self.log_level: LogLevel = None
        self.logger_types_map = LoggerTypesMap()

    def initialize(self, config: LoggingConfig=LoggingConfig()):

        for logger_type in self.logger_types:
            
            logger_name = logging_manager.logger_types.get_name(logger_type)
            logger_enabled = logging_manager.get_logger_enabled_state(logger_type)
            self.log_level = logging_manager.log_level

            config.from_dict({
                'logger_name': logger_name,
                'logger_enabled': logger_enabled,
                'logger_type': logger_type,
                'log_level': logging_manager.log_level,
                'logfiles_directory': logging_manager.logfiles_directory,
                'spinner_display': logging_manager.progress_display
            })

            ASYNC_TYPE = Type[AsyncLogger]
            SYNC_TYPE = Type[SyncLogger]
            
            if logger_type == LoggerTypes.SPINNER:
                ASYNC_TYPE = Type[AsyncSpinner]
                SYNC_TYPE = None

            elif logger_type == LoggerTypes.FILESYSTEM:
                ASYNC_TYPE = Type[AsyncFilesystemLogger]
                SYNC_TYPE = Type[SyncFilesystemLogger]

            logger: Logger[ASYNC_TYPE, SYNC_TYPE] = Logger(config)

            if logger_type == LoggerTypes.CONSOLE:
                logger.set_patterns('%(message)s')

            elif logger_type == LoggerTypes.DISTRIBUTED or logger_type == LoggerTypes.HEDRA:

                logger.set_patterns(
                    '%(asctime)s - %(name)s - %(levelname)s - %(module)s:%(funcName)s:%(lineno)d - %(message)s',
                    datefmt_pattern='%Y-%m-%dT%H:%M:%S.%Z'
                )

            self.loggers[logger_name] = logger

    def __getitem__(self, logger_name: str) -> Logger[Union[AsyncLogger, AsyncFilesystemLogger, AsyncSpinner], Union[SyncLogger, SyncFilesystemLogger]]:
        logger = self.loggers.get(logger_name)
        if logger is None:
            logger = Logger(
                logger_name,
                LoggerTypes.CONSOLE,
                log_level=logging_manager.log_level
            )

            logger.set_patterns('%(message)s')

        return logger

    @property
    def console(self) -> Logger[AsyncLogger, SyncLogger]:
        return self.loggers['console']

    @property
    def distributed(self) -> Logger[AsyncLogger, SyncLogger]:
        return self.loggers['distributed']

    @property
    def hedra(self) -> Logger[AsyncLogger, SyncLogger]:
        return self.loggers['hedra'] 

    @property
    def filesystem(self) -> Logger[AsyncFilesystemLogger, SyncFilesystemLogger]:
        return self.loggers['filesystem']

    @property
    def distributed_filesystem(self) -> Logger[AsyncFilesystemLogger, SyncFilesystemLogger]:
        return self.loggers['distributed_filesystem']

    @property
    def spinner(self) -> AsyncSpinner:
        return self.loggers['spinner'].aio

    def create_logger(
        self,
        logger_name: str,
        logger_type: LoggerTypes,
        rotation_interval_type: RolloverInterval=RolloverInterval.DAYS,
        rotation_interval: int=1,
        backups: int=1,
        rotation_time: datetime.time=None
    ):

        logger_enabled = logging_manager.get_logger_enabled_state(logger_type)

        ASYNC_TYPE = Type[AsyncLogger]
        SYNC_TYPE = Type[SyncLogger]

        if logger_type == LoggerTypes.FILESYSTEM:
            ASYNC_TYPE = Type[AsyncFilesystemLogger]
            SYNC_TYPE = Type[SyncFilesystemLogger]

        self.loggers[logger_name]: Logger[ASYNC_TYPE, SYNC_TYPE] = Logger(
            logger_name,
            logger_type,
            log_level=logging_manager.log_level,
            logger_enabled=logger_enabled,
            rotation_interval_type=rotation_interval_type,
            rotation_interval=rotation_interval,
            backup_count=backups,
            rotation_time=rotation_time
        )

    def disable_logger(self, logger_name):
        logger = self.loggers.get(logger_name)
        if logger:
            logger.logger_enabled = False

    def enable_logger(self, logger_name):
        logger = self.loggers.get(logger_name)
        if logger:
            logger.logger_enabled = True


import uuid
from playwright.async_api import Page
from hedra.tools.helpers import awaitable
from .command import PlaywrightCommand


class CommandLibrary:

    __slots__ = (
        'page'
    )


    def __init__(self, page) -> None:
        self.page: Page = page

    async def goto(self, command: PlaywrightCommand):
        await self.page.goto(
            command.url.location, 
            **command.options.extra
        )

    async def get_url(self, command: PlaywrightCommand):
        return self.page.url

    async def fill(self, command: PlaywrightCommand):
        await self.page.fill(
            command.page.selector, 
            command.input.text
        )

    async def check(self, command: PlaywrightCommand):
        await self.page.check(
            command.page.selector, 
            **command.options.extra
        )

    async def click(self, command: PlaywrightCommand):
        await self.page.click(
            command.page.selector, 
            **command.options.extra
        )

    async def double_click(self, command: PlaywrightCommand):
        await self.page.dblclick(
            command.page.selector, 
            **command.options.extra
        )

    async def submit_event(self, command: PlaywrightCommand):
        await self.page.dispatch_event(
            command.page.selector, 
            command.options.event, 
            **command.options.extra
        )

    async def drag_and_drop(self, command: PlaywrightCommand):
        await self.page.drag_and_drop(
            source_position=command.page.x_coordinate, 
            target_position=command.page.y_coordinate,
            **command.options.extra
        )

    async def switch_active_tab(self, command: PlaywrightCommand):
        await self.page.bring_to_front()

    async def evaluate_selector(self, command: PlaywrightCommand):
        await self.page.eval_on_selector(
            command.page.selector, 
            command.input.expression, 
            *command.input.args, 
            **command.options.extra
        )

    async def evaluate_all_selectors(self, command: PlaywrightCommand):
        await self.page.eval_on_selector_all(
            command.page.selector,
            command.input.expression, 
            *command.input.args, 
            **command.options.extra
        )

    async def evaluate_expression(self, command: PlaywrightCommand):
        await self.page.evaluate(
            command.input.expression, 
            *command.input.args, 
            **command.options.extra
        )

    async def evaluate_handle(self, command: PlaywrightCommand):
        await self.page.evaluate_handle(
            command.input.expression, 
            *command.input.args, 
            **command.options.extra
        )

    async def exepect_console_message(self, command: PlaywrightCommand):
        await self.page.expect_console_message(
            predicate=await self.page.on(
                'console',
                command.input.expression
            ), 
            **command.options.extra
        )

    async def expect_download(self, command: PlaywrightCommand):
        await self.page.expect_download(
            predicate=await self.page.on(
                'download', 
                command.input.expression
            ), 
            **command.options.extra
        )

    async def expect_event(self, command: PlaywrightCommand):
        await self.page.expect_event(
            command.options.event, 
            **command.options.extra
        )

    async def expect_location(self, command: PlaywrightCommand):
        await self.page.expect_navigation(
            url=command.url.location, 
            **command.options.extra
        )

    async def expect_popup(self, command: PlaywrightCommand):
        await self.page.expect_popup(
            predicate=await self.page.on(
                'popup', 
                command.input.expression
            )
        )

    async def expect_request(self, command: PlaywrightCommand):
        return await self.page.expect_request(
            command.url.location, 
            **command.options.extra
        )

    async def expect_request_finished(self, command: PlaywrightCommand):
        await self.page.expect_request_finished(
            await self.expect_request(command)
        )

    async def expect_response(self, command: PlaywrightCommand):
        await self.page.expect_response(
            command.url.location, 
            **command.options.extra
        )

    async def focus(self, command: PlaywrightCommand):
        await self.page.focus(
            command.page.selector, 
            **command.options.extra
        )

    async def hover(self, command: PlaywrightCommand):
        await self.page.hover(
            command.page.selector, 
            **command.options.extra
        )

    async def get_inner_html(self, command: PlaywrightCommand):
        return await self.page.inner_html(
            command.page.selector, 
            **command.options.extra
        )

    async def get_text(self, command: PlaywrightCommand):
        return await self.page.inner_text(
            command.page.selector, 
            **command.options.extra
        )

    async def get_input_value(self, command: PlaywrightCommand):
        return await self.page.inner_value(
            command.page.selector, 
            **command.options.extra
        )

    async def press_key(self, command: PlaywrightCommand):
        await self.page.press(
            command.page.selector,
            command.input.key, 
            **command.options.extra
        )

    async def verify_is_enabled(self, command: PlaywrightCommand):
        return await self.page.is_enabled(
            command.page.selector, 
            **command.options.extra
        )

    async def verify_is_hidden(self, command: PlaywrightCommand):
        return await self.page.is_hidden(
            command.page.selector, 
            **command.options.extra
        )

    async def verify_is_visible(self, command: PlaywrightCommand):
        return await self.page.is_hidden(
            command.page.selector, 
            **command.options.extra
        )

    async def verify_is_checked(self, command: PlaywrightCommand):
        return await self.page.is_checked(
            command.page.selector, 
            **command.options.extra
        )

    async def get_content(self, command: PlaywrightCommand):
        return await self.page.content()

    async def get_element(self, command: PlaywrightCommand):
        return await self.page.query_selector(
            command.page.selector, 
            **command.options.extra
        )

    async def get_all_elements(self, command: PlaywrightCommand):
        return await self.page.query_selector_all(
            command.page.selector, 
            **command.options.extra
        )

    async def reload_page(self, command: PlaywrightCommand):
        await self.page.reload(**command.options.extra)

    async def take_screenshot(self, command: PlaywrightCommand):
        await self.page.screenshot(
            path=f'{command.input.path}.png', 
            **command.options.extra
        )

    async def select_option(self, command: PlaywrightCommand):

        if command.input.by_label:
            await self.page.select_option(
                command.page.selector, 
                label=command.input.option, 
                **command.options.extra
            )

        elif command.input.by_value:
            await self.page.select_option(
                command.page.selector, 
                value=command.input.option, 
                **command.options.extra
            )

        else:
            await self.page.select_option(
                command.page.selector, 
                command.input.option, 
                **command.options.extra
            )

    async def set_checked(self, command: PlaywrightCommand):
        await self.page.set_checked(
            command.page.selector, 
            command.options.is_checked, 
            **command.options.extra
        )

    async def set_default_timeout(self, command: PlaywrightCommand):
        await self.page.set_default_timeout(command.options.timeout)

    async def set_navigation_timeout(self, command: PlaywrightCommand):
        await self.page.set_default_navigation_timeout(command.options.timeout)

    async def set_http_headers(self, command: PlaywrightCommand):
        await self.page.set_extra_http_headers(command.url.headers)

    async def tap(self, command: PlaywrightCommand):
        await self.page.tap(
            command.page.selector, 
            **command.options.extra
        )

    async def get_text_content(self, command: PlaywrightCommand):
        await self.page.text_content(
            command.page.selector,
            **command.options.extra
        )

    async def get_page_title(self, command: PlaywrightCommand):
        await self.page.title()

    async def input_text(self, command: PlaywrightCommand):
        await self.page.type(
            command.page.selector,
            command.input.text, 
            **command.options.extra
        )

    async def uncheck(self, command: PlaywrightCommand):
        await self.page.uncheck(
            command.page.selector, 
            **command.options.extra
        )

    async def wait_for_event(self, command: PlaywrightCommand):
        await self.page.wait_for_event(
            command.options.event, 
            **command.options.extra
        )

    async def wait_for_function(self, command: PlaywrightCommand):
        await self.page.wait_for_function(
            command.input.expression, 
            *command.input.args, 
            **command.options.extra
        )

    async def wait_for_page_load_state(self, command: PlaywrightCommand):
        await self.page.wait_for_load_state(**command.options.extra)

    async def wait_for_selector(self, command: PlaywrightCommand):
        await self.page.wait_for_selector(
            command.page.selector, 
            **command.options.extra
        )

    async def wait_for_timeout(self, command: PlaywrightCommand):
        await self.page.wait_for_timeout(command.options.timeout)

    async def wait_for_url(self, command: PlaywrightCommand):
        await self.page.wait_for_url(
            command.url.location, 
            **command.options.extra
        )

    async def switch_frame(self, command: PlaywrightCommand):

        if command.url.location == 'url':
            await self.page.frame(url=command.url.location)

        else:
            await self.page.frame(name=command.page.frame)

    async def get_frames(self, command: PlaywrightCommand):
        return await self.page.frames()

    async def get_attribute(self, command: PlaywrightCommand):
        return await self.page.get_attribute(
            command.page.selector,
            command.page.attribute
        )

    async def go_back_page(self, command: PlaywrightCommand):
        await self.page.go_back(**command.options.extra)

    async def go_forward_page(self, command: PlaywrightCommand):
        await self.page.go_forward(**command.options.extra)


from .graphql import MercuryGraphQLClient
from .grpc import MercuryGRPCClient
from .graphql_http2 import MercuryGraphQLHTTP2Client
from .http import MercuryHTTPClient
from .http2 import MercuryHTTP2Client
from .playwright import MercuryPlaywrightClient
from .websocket import MercuryWebsocketClient
from .udp import MercuryUDPClient
from .task import MercuryTaskRunner


from .action import action
from .after import after
from .before import before
from .channel import channel
from .check import check
from .context import context
from .depends import depends
from .event import event
from .internal import Internal  
from .metric import metric
from .restore import restore
from .save import save
from .setup import setup
from .task import task
from .teardown import teardown
from .validate import validate


import binascii
from .base_event import BaseEvent



class ConnectionTerminated(BaseEvent):
    event_type='CONNECTION_TERMINATED'

    __slots__ = (
        'error_code',
        'last_stream_id',
        'additional_data'
    )

    """
    The ConnectionTerminated event is fired when a connection is torn down by
    the remote peer using a GOAWAY frame. Once received, no further action may
    be taken on the connection: a new connection must be established.
    """
    def __init__(self):
        #: The error code cited when tearing down the connection. Should be
        #: one of :class:`ErrorCodes <h2.errors.ErrorCodes>`, but may not be if
        #: unknown HTTP/2 extensions are being used.
        self.error_code = None

        #: The stream ID of the last stream the remote peer saw. This can
        #: provide an indication of what data, if any, never reached the remote
        #: peer and so can safely be resent.
        self.last_stream_id = None

        #: Additional debug data that can be appended to GOAWAY frame.
        self.additional_data = None

    def __repr__(self):
        additional_data = b''
        if self.additional_data:
            additional_data = binascii.hexlify(self.additional_data[:20]).decode('ascii')
        return (
            "<ConnectionTerminated error_code:%s, last_stream_id:%s, "
            "additional_data:%s>" % (
                self.error_code,
                self.last_stream_id,
                additional_data
        ))



import asyncio
import time
from hedra.runners.leader_services.job_registry import job
from .job import Job

from hedra.runners.worker_services.pipelines import WorkerPipeline
from hedra.runners.worker_services.leader_registry import LeaderServicesManager


class JobRegistry:

    def __init__(self, config, reporter_config, worker_address, worker_id) -> None:
        self.config = config
        self.reporter_config = reporter_config
        self.jobs = {}
        self.created = {}
        self.active = {}
        self.completed = {}
        self.failed = {}

        self.monitor = None
        self.monitoring = False

        worker_ip, worker_port = worker_address.split(':')
        worker_port = int(worker_port)

        self.leaders = LeaderServicesManager(config, worker_ip, worker_port, worker_id)
        
        self.worker_address = worker_address
        self.worker_id = worker_id

        logger = Logger()
        self.session_logger = logger.generate_logger()

        self.heartbeat_interval = config.distributed_config.get('heartbeat_interval', 1)

    async def __iter__(self):
        for job_id in self.jobs:
            yield self.jobs[job_id]

    async def __aiter__(self):
        for job_id in self.jobs:
            yield self.jobs[job_id]

    def __getitem__(self, job_id):
        if job_id in self.jobs:
            return self.jobs[job_id]

        return None

    async def iter_created(self):
        for job_id in self.created:
            yield self.created[job_id]

    async def iter_active(self):
        for job_id in self.active:
            yield self.active[job_id]

    async def iter_completed(self):
        for job_id in self.completed:
            yield self.completed[job_id]

    async def iter_failed(self):
        for job_id in self.completed:
            yield self.completed[job_id]

    async def get_from_active(self, job_id):
        if job_id in self.active:
            return self.active[job_id]

        return None

    async def create_job(self, job_id, new_job, leader_address, timeout=300):

        if job_id not in self.created or job_id not in self.active:

            pipeline = WorkerPipeline(
                self.config,
                worker_address=self.worker_address,
                worker_id=self.worker_id,
                timeout=timeout
            )

            leaders = await self.leaders.registered()
            await pipeline.setup_pipeline(
                new_job,
                self.reporter_config,
                leaders
            )

            job = Job(job_id, pipeline)

            await job.attach_leader_to_job(leader_address)
            self.session_logger.info(f'\nAttaching leader - {leader_address} to pending job id - {job_id}\n')


            self.created[job_id] = job
            self.jobs[job_id] = job

            await self.created[job_id].set_job_status('created')

        else:
            job = self.jobs[job_id]

            if leader_address not in job.registered_leaders:
                await job.attach_leader_to_job(leader_address)
                self.session_logger.info(f'Attaching leader - {leader_address} to pending job id - {job_id}\n')

    async def start_job(self, job_id):
        if job_id not in self.active:
            self.session_logger.info(f'Starting job - {job_id}.')

            job = self.created[job_id]

            await job.run()
            await job.start_timer()

            self.active[job_id] = job

            await self.active[job_id].set_job_status('active')

            del self.created[job_id]

        await self.jobs[job_id].set_job_status('active')

    async def update_job_leaders(self):
        active_leaders = await self.leaders.registered()
        for job_id in self.active:
            active_job = self.active[job_id]

            await active_job.pipeline.update_leaders(
                active_leaders,
                self.leaders.registry.failed,
                self.leaders.registry.registered_leaders
            )

    async def prune_expired_jobs(self):
        active = {}

        for job_id in self.active:
            active_job = self.active[job_id]
            await active_job.update_time_elapsed()
            expired = await active_job.check_if_deadline_exceeded()
            
            if expired:
                await active_job.set_job_status('failed')
                self.failed[job_id] = active_job
                self.jobs[job_id] = active_job

            else:
                active[job_id] = active_job
    
        self.active = active

    async def start_job_monitor(self):
        self.monitoring = True
        self.monitor = asyncio.create_task(
            self._monitor_jobs()
        )

    async def _monitor_jobs(self):
        while self.monitoring:
            await self.prune_expired_jobs()
            await asyncio.sleep(self.heartbeat_interval)
            self.failed = {}

    async def stop_job_monitor(self):
        self.monitoring = False
        await self.monitor

    async def set_job_complete(self, job_id):
        active = dict(self.active)
        not_completed = {}

        job = active[job_id]

        job_completed = await job.check_if_completed()
        if job_completed:
            await job.set_job_status('completed')
            self.completed[job_id] = job
            await self.jobs[job_id].set_job_status('completed')

        else:
            not_completed[job_id] = job

        self.active = not_completed

    async def wait_for_leaders(self):
        start = time.time()
        elapsed = 0
        leaders_registered = await self.leaders.registry.registered_count_equals_expected_count()
        while leaders_registered is False and elapsed < self.leaders.leader_request_timeout:
            leaders_registered = await self.leaders.registry.registered_count_equals_expected_count()
            elapsed = time.time() - start
            await asyncio.sleep(1)



from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import ValidateHook
from hedra.core.graphs.stages.validate.exceptions import HookValidationError
from .base_hook_validator import BaseHookVaidator


class ValidateHookValidator(BaseHookVaidator):

    def __init__(self, metadata_string: str) -> None:
        super(
            ValidateHookValidator,
            self
        ).__init__(metadata_string)

    async def validate(self, hook: ValidateHook):

        try:

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validating {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

            assert hook.hook_type is HookType.VALIDATE, f"Hook type mismatch - hook {hook.name}:{hook.hook_id} is a {hook.hook_type.name} hook, but Hedra expected a {HookType.VALIDATE.name} hook."
            assert hook.shortname in hook.name, f"Shortname {hook.shortname} must be contained in full Hook name {hook.name}:{hook.hook_id} for @validate hook {hook.name}:{hook.hook_id}."
            assert hook.call is not None, f"Method is not not found on stage or was not supplied to hook - {hook.name}:{hook.hook_id}"
            assert hook.call.__code__.co_argcount > 1, f"Missing required argument 'hook' for @validate hook {hook.name}:{hook.hook_id}"
            assert hook.call.__code__.co_argcount < 3, f"Too many args. - @validate hook {hook.name}:{hook.hook_id} only requires 'events_group' as additional arg."
            assert hook.stage is not None, f"No target stage name provided for @validate hook {hook.name}:{hook.hook_id}. Please specify the stage name whose hooks you wish to validate."
            assert len(hook.names) > 0, f"No target hook names provided for @validate hook {hook.name}:{hook.hook_id}. Please specify at least one hook to validate."
            assert 'self' in hook.call.__code__.co_varnames

            for name in hook.names:
                hook_for_validation = self.hooks_by_name.get(name)
                stage, hook_name = name.split('.')

                assert hook_for_validation is not None, f"Specified hook {hook_name} for stage {stage} not found for @validate hook {hook.name}:{hook.hook_id}"

                await hook.call(hook_for_validation.call)

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validated {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

            
        except AssertionError as hook_validation_error:
                raise HookValidationError(hook.stage_instance, str(hook_validation_error))


import asyncio
from .bootstrap_client import BoostrapClient


class BootstrapManager:

    def __init__(self, config) -> None:
        self.client = BoostrapClient(config)
        self.use_bootstrap = self.client.use_bootstrap

    async def register_leader(self, leader_address):
        await self.client.register_self(leader_address)

    async def wait_for_leaders(self, leader_address, leader_id):
        while self.client.registry.ready is False:
            await self.client.get_registration_status(leader_address, leader_id)
            await asyncio.sleep(1)

    async def discovered(self):
        return list(self.client.registry.discovered.values())



import os
from typing import List
from git.repo import Repo
from .action import RepoAction
from .config import RepoConfig


class Initialize(RepoAction):

    def __init__(self, config: RepoConfig, graph_files: List[str]) -> None:
        super(
            Initialize,
            self
        ).__init__(config, graph_files)

    def execute(self):

        if os.path.exists(self.config.path) is False:
            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Creating repo directory at path - {self.config.path}')
            os.makedirs(self.config.path)

        try:
            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Re-initializing repo')
            self._setup()

            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Updating .gitignore')
            self._update_ignore()

            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Setting and fetching new branches from remote')
            self.remote = self.repo.remote(self.config.remote)
            self.remote.fetch()

            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Checking out branch - {self.config.branch} - and updating remote tracking')
            self._checkout()

        except Exception:

            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Initializing new repository at - {self.config.path}')
            self.repo = Repo.init(self.config.path)

            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Updating .gitignore')
            self._update_ignore()

            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Adding {len(self.discovered_files)} to new repo in commit')
            self.repo.index.add(self.discovered_files)
            self.repo.index.commit(f'Initialized new graph repo at - {self.config.path}')


            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Setting remote as - {self.config.remote} - with URL - {self.config.uri}')
            remote_names = [
                remote.name for remote in self.repo.remotes
            ]
            if self.config.remote not in remote_names:
                self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Creating new remote as - {self.config.remote}')
                self.repo.create_remote(self.config.remote, self.config.uri)

            self.remote = self.repo.remote(self.config.remote)
            self.remote.set_url(self.config.uri)
        
            self.logger.hedra.sync.debug(f'InitializeAction: {self.action_id} - Checking out branch - {self.config.branch} - and updating remote tracking')
            self._checkout()

        



from .action_hook_validator import ActionHookValidator
from .after_hook_validator import AfterHookValidator
from .base_hook_validator import BaseHookVaidator
from .before_hook_validator import BeforeHookVaidator
from .channel_hook_validator import ChannelHookValidator
from .check_hook_validator import CheckHookValidator
from .event_hook_validator import EventHookVaidator
from .metric_hook_validator import MetricHookVaidator
from .restore_hook_validator import RestoreHookValidator
from .save_hook_validator import SaveHookValidator
from .setup_hook_validator import SetupHookVaidator
from .task_hook_validator import TaskHookValidator
from .teardown_hook_validator import TeardownHookValidator
from .validate_hook_validator import ValidateHookValidator



from .base_event import BaseEvent


class WindowUpdated(BaseEvent):
    event_type='WINDOW_UPDATED'
    __slots__ = ('stream_id', 'delta')
    """
    The WindowUpdated event is fired whenever a flow control window changes
    size. HTTP/2 defines flow control windows for connections and streams: this
    event fires for both connections and streams. The event carries the ID of
    the stream to which it applies (set to zero if the window update applies to
    the connection), and the delta in the window size.
    """
    def __init__(self):
        #: The Stream ID of the stream whose flow control window was changed.
        #: May be ``0`` if the connection window was changed.
        self.stream_id = None

        #: The window delta.
        self.delta = None

    def __repr__(self):
        return "<WindowUpdated stream_id:%s, delta:%s>" % (
            self.stream_id, self.delta
        )


from enum import Enum


class HookType(Enum):
    ACTION='ACTION'
    SETUP='SETUP'
    TEARDOWN='TEARDOWN'
    BEFORE='BEFORE'
    AFTER='AFTER'
    CHECK='CHECK'
    METRIC='METRIC'
    VALIDATE='VALIDATE'
    INTERNAL='INTERNAL'
    SAVE='SAVE'
    TASK='TASK'
    CHANNEL='CHANNEL'
    EVENT='EVENT'
    CONTEXT='CONTEXT'
    RESTORE='RESTORE'



from typing import List
from hedra.core.engines.types.common.types import RequestTypes
from .types import (
    GraphQLEvent,
    GRPCEvent,
    HTTPEvent,
    HTTP2Event,
    PlaywrightEvent,
    TaskEvent,
    UDPEvent,
    WebsocketEvent,
)



results_types = {
    RequestTypes.GRAPHQL: GraphQLEvent,
    RequestTypes.GRPC: GRPCEvent,
    RequestTypes.HTTP: HTTPEvent,
    RequestTypes.HTTP2: HTTP2Event,
    RequestTypes.PLAYWRIGHT: PlaywrightEvent,
    RequestTypes.TASK: TaskEvent,
    RequestTypes.UDP: UDPEvent,
    RequestTypes.WEBSOCKET: WebsocketEvent
}




from typing import Dict, Iterator, Union, List
from hedra.core.engines.types.common.hooks import Hooks
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http.action import HTTPAction
from .utils import (
    pack_hostname,
    create_sec_websocket_key
)
from .constants import WEBSOCKETS_VERSION


class WebsocketAction(HTTPAction):

    def __init__(
        self,
        name: str, 
        url: str, 
        method: str = 'GET', 
        headers: Dict[str, str] = {}, 
        data: Union[str, dict, Iterator, bytes, None] = None, 
        user: str=None, 
        tags: List[Dict[str, str]] = []
    ) -> None:

        super(
            WebsocketAction,
            self
        ).__init__(
            name, 
            url, 
            method, 
            headers, 
            data, 
            user, 
            tags
        )

        self.type = RequestTypes.WEBSOCKET
        self.hooks: Hooks[WebsocketAction] = Hooks()

    def _setup_headers(self):

        for header_name, header_value in self._headers.items():
            header_name_lowered = header_name.lower()
            self._headers[header_name_lowered] = header_value

        headers = [
            "GET %s HTTP/1.1" % self.url.path,
            "Upgrade: websocket"
        ]
        if self.url.port == 80 or self.url.port == 443:
            hostport = pack_hostname(self.url.hostname)
        else:
            hostport = "%s:%d" % (pack_hostname(
                self.url.hostname
            ), self.url.port)

        host = self._headers.get("host")
        if host:
            headers.append(f"Host: {host}")
        else:
            headers.append(f"Host: {hostport}")

        # scheme, url = url.split(":", 1)
        if  not self._headers.get("suppress_origin"):

            origin = self._headers.get("origin")

            if origin:
                headers.append(f"Origin: {origin}")

            elif self.url.scheme == "wss":
                headers.append(f"Origin: https://{hostport}")
                
            else:
                headers.append(f"Origin: http://{hostport}")

        key = create_sec_websocket_key()

        header = self._headers.get("header")
        if not header or 'Sec-WebSocket-Key' not in header:
            headers.append(f"Sec-WebSocket-Key: {key}")
        else:
            key = self._headers.get("header", {}).get('Sec-WebSocket-Key')

        if not header or 'Sec-WebSocket-Version' not in header:
            headers.append(f"Sec-WebSocket-Version: {WEBSOCKETS_VERSION}")

        connection = self._headers.get('connection')
        if not connection:
            headers.append('Connection: Upgrade')
        else:
            headers.append(connection)

        subprotocols = self._headers.get("subprotocols")
        if subprotocols:
            headers.append("Sec-WebSocket-Protocol: %s" % ",".join(subprotocols))

        if header:
            if isinstance(header, dict):
                header = [
                    ": ".join([k, v])
                    for k, v in header.items()
                    if v is not None
                ]
            headers.extend(header)

        headers.append("")
        headers.append("")

        self.encoded_headers = '\r\n'.join(headers).encode()


from typing import Dict
from collections import defaultdict
from hedra.core.graphs.hooks.registry.registry_types.hook import Hook
from hedra.logging import HedraLogger



class BaseHookVaidator:

    def __init__(self, metadata_string: str) -> None:
        self.logger = HedraLogger()
        self.logger.initialize()
        self.metadata_string: str = metadata_string
        self.hooks_by_name: Dict[str, Hook] = defaultdict(dict)

    async def validate(self, hook: Hook):
        pass



import uuid
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    from hedra.reporting.types.statsd import StatsD
    from aio_statsd import TelegrafClient
    from .telegraf_config import TelegrafConfig
    has_connector = True

except Exception:
    from hedra.reporting.types.empty import Empty as StatsD
    TelegrafConfig=None
    has_connector = False


class Telegraf(StatsD):

    def __init__(self, config: TelegrafConfig) -> None:
        self.host = config.host
        self.port = config.port

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()


        self.connection = TelegrafClient(
            host=self.host,
            port=self.port
        )

        self.statsd_type = 'Telegraf'

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to {self.statsd_type}')

        for event in events:
            self.connection.send_telegraf(event.name, {'time': event.time})
            
            if event.success:
                self.connection.send_telegraf(event.name, {'success': 1})
            
            else:
                self.connection.send_telegraf(event.name, {'failed': 1})

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to {self.statsd_type}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            self.connection.send_telegraf(
                f'{metrics_set.name}_common',
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to {self.statsd_type}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to {self.statsd_type}')

        for metrics_set in metrics:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')


            for group_name, group in metrics_set.groups.items():
                self.connection.send_telegraf(
                    f'{metrics_set.name}_{group_name}', 
                    {
                        **group.record,
                        'group': group_name
                    }
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to {self.statsd_type}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():
                self.connection.send_telegraf(
                    f'{metrics_set.name}_{custom_group_name}',
                    {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': custom_group_name,
                        **group
                    }
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to {self.statsd_type}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                self.connection.send_telegraf(
                    f'{metrics_set.name}_errors', {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'error_message': error.get('message'),
                        'error_count': error.get('count')
                    })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to {self.statsd_type}')


import asyncio
import uuid
from .bootstrap_server import BootstrapServer


class BootstrapManager:

    def __init__(self, config) -> None:
        self.leader_id = uuid.uuid4()

        self.session_logger.info('Initializing bootstrap server...')

        self.server = BootstrapServer(config, self.leader_id)

    async def start_server(self):
        self.session_logger.info(f'Starting bootstrap at - {self.server.service.bootstrap_service_address}\n')
        await self.server.run()
        await self.server.service.broadkast_server.start()
        self.server.run_server = True

    async def wait(self):
        while self.server.run_server:
            await asyncio.sleep(1)

    async def stop(self):
        await self.server.stop_server()
        await self.server.service.broadkast_server.stop()




from typing import List
from hedra.plugins.types.common import Event
from hedra.plugins.types.reporter import (
    ReporterConfig,
    ReporterPlugin,
    Metrics,
    reporter_connect,
    reporter_close,
    process_events,
    process_shared,
    process_metrics,
    process_custom,
    process_errors
)


class CustomReporterConfig(ReporterConfig):
    pass


class  CustomReporter(ReporterPlugin):
    config=CustomReporterConfig

    def __init__(self, config: CustomReporterConfig) -> None:
        super().__init__(config)

    @reporter_connect()
    async def reporter_connect(self):
        pass

    @process_events()
    async def reporter_process_events(self, events: List[Event]):
        pass

    @process_shared()
    async def reporter_process_shared_metrics(self, metrics: List[Metrics]):
        pass

    @process_metrics()
    async def reporter_process_metrics(self, metrics: List[Metrics]):
        pass

    @process_custom()
    async def reporter_process_custom_metrics(self, metrics: List[Metrics]):
        pass

    @process_errors()
    async def reporter_process_errors(self, metrics: List[Metrics]):
        pass

    @reporter_close()
    async def reporter_close(self):
        pass


from typing import Coroutine, Dict, List, Union
from .types import RequestTypes


class BaseResult:

    __slots__ = (
        'action_id',
        'name',
        'checks',
        'error',
        'source',
        'user',
        'tags',
        'type',
        'time',
        'wait_start',
        'start',
        'connect_end',
        'write_end',
        'complete'
    )

    def __init__(
        self, 
        action_id: str,
        name: str, 
        source: str,
        user: str,
        tags: List[Dict[str, str]],
        type: Union[RequestTypes, str],
        checks: List[Coroutine], 
        error: Exception
    ) -> None:
        self.action_id = action_id
        self.name = name
        self.checks = checks
        self.error = error
        self.source = source
        self.user = user
        self.tags = tags
        self.type = type

        self.time = 0
        self.wait_start = 0
        self.start = 0
        self.connect_end = 0
        self.write_end = 0
        self.complete = 0

    def to_dict(self):
        return {
            'name': self.name,
            'error': str(self.error),
            'source': self.source,
            'user': self.user,
            'tags': self.tags,
            'type': self.type,
            'wait_start': float(self.wait_start),
            'start': float(self.start),
            'connect_end': float(self.connect_end),
            'write_end': float(self.write_end),
            'complete': float(self.complete)
        }


import grpc
import os
import psutil
import asyncio
from concurrent import futures

from .service import WorkerService
from .proto import add_WorkerServerServicer_to_server

class WorkerServer:

    def __init__(self, config, reporter_config, worker_id):
        logger = Logger()
        self.loop = None
        self.worker_id = worker_id

        self.session_logger = logger.generate_logger('hedra')
        self.config = config.distributed_config
        self.worker_ip = self.config.get(
            'worker_ip',
            os.getenv(
                'WORKER_IP', 
                'localhost'
            )
        )
        self.worker_port = self.config.get(
            'worker_server_port',
            os.getenv('WORKER_SERVER_PORT',  6671)
        )
        self.address = config.distributed_config.get(
            'worker_server_address',
            os.getenv('WORKER_SERVER_ADDRESS')
        )

        if self.address is None:
            self.address = f'{self.worker_ip}:{self.worker_port}'

        leader_ips = config.distributed_config.get(
            'leader_ips',
            os.getenv('LEADER_IPS', 'localhost')
        )

        leader_ports = config.distributed_config.get(
            'leader_ports',
            os.getenv('LEADER_PORTS', "6669")
        )

        leader_addresses = config.distributed_config.get(
            'leader_addresses',
            os.getenv('LEADER_ADDRESSES')
        )

        if leader_addresses:
            leader_addresses = leader_addresses.split(",")
            self.leader_addresses = leader_addresses

        else:
            leader_ips = leader_ips.split(",")
            leader_ports = [int(port) for port in leader_ports.split(",")]

            self.leader_addresses = []

            for leader_ip, leader_port in zip(leader_ips, leader_ports):
                self.leader_addresses.append(
                    f'{leader_ip}:{leader_port}'
                )

        self.session_logger.info('Initializing updates server...')
        self._worker_threads = self.config.get(
            'max_workers', 
            os.getenv(
                'MAX_THREAD_WORKERS',
                psutil.cpu_count(logical=False)
            )
        )

        self.session_logger.debug('Generating server with - {threads} - threads'.format(threads=self._worker_threads))
        self.server = grpc.aio.server()

        self.service = WorkerService(
            config,
            reporter_config,
            self.address,
            self.worker_id
        )

        add_WorkerServerServicer_to_server(
            self.service,
            self.server
        )

        server_address = '[::]:{port}'.format(port=self.worker_port)
        self.server.add_insecure_port(server_address)

        self.run_server = False
        self.loop = None

    async def run(self):
        self.session_logger.info('Updates server serving on port - {port}'.format(port=self.worker_port))
        
        leader_address_strings = ''.join([f'\n- {leader_address}' for leader_address in self.leader_addresses])
        
        self.session_logger.info(f'\nUpdates server connecting to leaders at addresses: {leader_address_strings}\n')

        await self.service.job_registry.leaders.initialize()
        await self.server.start()
        self.service.running = True
        self.run_server = True

    async def stop_server(self):
        await self.server.stop(0)

    async def add_new_leaders(self, new_leaders_queue):
        await self.service.job_registry.leaders.add_new_leaders(new_leaders_queue)

    async def register_worker(self):
        await self.service.job_registry.leaders.register_worker()

    async def start_monitor(self):
        await self.service.job_registry.start_job_monitor()

    async def stop_monitor(self):
        await self.service.job_registry.stop_job_monitor()




import json
from typing import Dict, Any, Union
from hedra.core.engines import registered_engines
from hedra.core.engines.client.config import Config
from hedra.core.graphs.hooks.registry.registrar import registrar
from hedra.core.engines.types.graphql import GraphQLAction, MercuryGraphQLClient
from hedra.core.graphs.hooks.registry.registry_types import ActionHook, TaskHook
from hedra.core.personas.types.default_persona import DefaultPersona
from hedra.core.engines.types.common.types import RequestTypes
from hedra.logging import HedraLogger


async def assemble_graphql_action(
    hook: Union[ActionHook, TaskHook],
    hook_action: Dict[str, Any],
    persona: DefaultPersona,
    config: Config,
    metadata_string: str
):

    logger = HedraLogger()
    logger.initialize()
    
    action_name = hook_action.get('name')
    hook_type_name = RequestTypes.GRAPHQL.capitalize()

    await logger.filesystem.aio['hedra.core'].debug(
        f'{metadata_string} - Assembling {hook_type_name} Hook - {action_name}'
    )
    
    action_hooks = hook_action.get('hooks', {})
    action_hook_names = action_hooks.get('names')

    before_hook_name = action_hook_names.get('before')
    after_hook_name = action_hook_names.get('after')
    check_hook_names = action_hook_names.get('checks', [])
    listener_names = action_hook_names.get('listeners', [])
    channel_hook_names = action_hook_names.get('channels', [])

    if before_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Before Hook - {before_hook_name}'
        )

    if after_hook_name:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found After Hook - {after_hook_name}'
        )

    for check_hook_name in check_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Check Hook - {check_hook_name}'
        )

    for channel_hook_name in channel_hook_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Channel Hook - {channel_hook_name}'
        )

    for listener_name in listener_names:
        await logger.filesystem.aio['hedra.core'].debug(
            f'{metadata_string} - {hook_type_name} Hook - {action_name} - Found Listener - {listener_name}'
        )

    hook.session: MercuryGraphQLClient = registered_engines.get(RequestTypes.GRAPHQL)(
        concurrency=persona.batch.size,
        timeouts=hook_action.get('timeouts'),
        reset_connections=hook_action.get('reset_connections')
    )

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Connect Timeout: {hook.session.timeouts.connect_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Read Timeout: {hook.session.timeouts.socket_read_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Total Timeout: {hook.session.timeouts.total_timeout}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Reset Connections: {hook.session.pool.reset_connections}')

    action_url = hook_action.get('url', {})
    action_headers = hook_action.get('headers', {})
    action_data = hook_action.get('data', {})
    action_metadata = hook_action.get('metadata', {})

    hook.action: GraphQLAction = GraphQLAction(
        action_name,
        action_url.get('url'),
        hook_action.get('method'),
        headers=action_headers.get('headers'),
        data=action_data.get('data'),
        user=action_metadata.get('user'),
        tags=action_metadata.get('tags')
    )

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - URL: {hook.action.url.full}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Method: {hook.action.method}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Headers: {json.dumps(hook.action.headers)}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Data: {json.dumps(hook.action.data)}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - User: {hook.action.metadata.user}')

    for tag in hook.action.metadata.tags_to_string_list():
        await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Tag: {tag}')

    hook.action.hooks.notify = action_hooks.get('notify', False)
    hook.action.hooks.listen = action_hooks.get('listen', False)


    if before_hook_name:
        before_hook = registrar.all.get(before_hook_name)
        hook.action.hooks.before = before_hook.call

    if after_hook_name:
        after_hook = registrar.all.get(after_hook_name)
        hook.action.hooks.after = after_hook.call

    hook.action.hooks.checks = []
    hook.action.hooks.checks.extend(check_hook_names)

    hook.action.hooks.channels = []
    for channel_hook_name in channel_hook_names:
        channel = registrar.all.get(channel_hook_name)
        hook.action.hooks.channels.append(channel.call) 

    hook.action.hooks.listeners = listener_names

    hook.action.url.ip_addr = action_url.get('ip_addr')
    hook.action.url.port = action_url.get('port')
    hook.action.url.socket_config = action_url.get('socket_config')
    hook.action.url.is_ssl = action_url.get('is_ssl')

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - IP Addr: {hook.action.url.ip_addr}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Port: {hook.action.url.port}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Socket Family: {hook.action.url.family}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Socket Protocol: {hook.action.url.protocol}')
    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Is SSL: {hook.action.url.is_ssl}')

    if hook.action.url.is_ssl:
        hook.action.ssl_context = hook.session.ssl_context

    hook.action.encoded_headers = action_headers.get('encoded_headers')
    hook.action.encoded_data = action_data.get('encoded_data')

    await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - {hook_type_name} Hook - {action_name} - Setup complete')

    return hook


from typing import Optional
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class NewRelicConfig(BaseModel):
    config_path: str
    environment: Optional[str]=None
    registration_timeout: int=60
    shutdown_timeout: int=60
    newrelic_application_name: str='hedra'
    reporter_type: ReporterTypes=ReporterTypes.NewRelic


import asyncio
from typing import Dict, Iterator, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.udp import UDPAction, UDPResult, MercuryUDPClient
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.logging import HedraLogger
from .base_client import BaseClient


class UDPClient(BaseClient[MercuryUDPClient, UDPAction, UDPResult]):

    def __init__(self, config: Config) -> None:
        super().__init__()

        self.session = MercuryUDPClient(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections
        )
        self.request_type = RequestTypes.UDP
        self.client_type = self.request_type.capitalize()

        self.next_name = None
        self.intercept = False
        self.waiter = None
        self.actions: ActionsStore = None
        self.registered = {}

        self.logger = HedraLogger()
        self.logger.initialize()


    def __getitem__(self, key: str):
        return self.session.registered.get(key)

    async def receive(
        self,
        url: str, 
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = UDPAction(
            self.next_name,
            url,
            wait_for_response=True,
            data=None,
            user=user,
            tags=tags             
        )

        return await self._execute_action(request)

    async def send(
        self,
        url: str, 
        wait_for_resonse: bool = False,
        data: Union[dict, str, bytes, Iterator] = None,
        user: str = None,
        tags: List[Dict[str, str]] = []
    ):

        request = UDPAction(
            self.next_name,
            url,
            wait_for_response=wait_for_resonse,
            data=data,
            user=user,
            tags=tags           
        )

        return await self._execute_action(request)

    


import asyncio
import time
import uuid
from typing import Any, Awaitable, Dict, Generic, TypeVar, Union
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.plugins.types.engine.action import Action
from hedra.plugins.types.engine.result import Result
from hedra.plugins.types.common.types import PluginHooks
from .connection import CustomConnection
from .pool import CustomPool

A = TypeVar('A')
R = TypeVar('R')

class MercuryCustomClient(Generic[A, R]):

    __slots__ = (
        'session_id',
        'timeouts',
        'registered',
        '_hosts',
        'closed',
        'sem',
        'active',
        'waiter',
        'plugin',
        '_on_connect',
        '_on_execute',
        '_on_close',
        'custom_connection',
        'pool'
    )

    def __init__(
        self, 
        plugin: Any,
        concurrency: int=10**3, 
        timeouts: Timeouts = Timeouts(), 
        reset_connections: bool=False
    ) -> None:

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, CustomConnection] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        
        self.active = 0
        self.waiter: asyncio.Future = None
        self.plugin: Any = plugin

        self._on_connect = self.plugin.hooks.get(PluginHooks.ON_ENGINE_CONNECT)
        self._on_execute = self.plugin.hooks.get(PluginHooks.ON_ENGINE_EXECUTE)
        self._on_close = self.plugin.hooks.get(PluginHooks.ON_ENGINE_CLOSE)

        self.custom_connection: CustomConnection = lambda reset_connection: CustomConnection(
            security_context=self.plugin.security_context,
            reset_connection=reset_connection,
            on_connect=self._on_connect.call,
            on_execute=self._on_execute.call,
            on_close=self._on_close.call
        )

        self.pool = CustomPool(
            self.custom_connection,
            concurrency, 
            reset_connections=reset_connections
        )

        self.pool.create_pool()


    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                self.custom_connection(self.pool.reset_connections)
            )
        
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = Semaphore(self.pool.size)

    async def wait_for_active_threshold(self):
        if self.waiter is None:
            self.waiter = asyncio.get_event_loop().create_future()
            await self.waiter

    async def prepare(self, action: Action) -> Awaitable[Union[CustomConnection, Exception]]:
        try:
            connection: CustomConnection = self.custom_connection(self.pool.reset_connections)

            if action.use_security_context:
                action.security_context = connection.security_context

            await asyncio.wait_for(
                connection.make_connection(action),
                timeout=self.timeouts.connect_timeout
            )

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

        except Exception as e:       
            raise e

    async def execute_prepared_request(self, action: Action) -> Awaitable[Result[R]]:
 
        result: Result = self.plugin.result(action)
        
        result.times['wait_start'] = time.monotonic()
        self.active += 1
 
        async with self.sem:
            connection = self.pool.connections.pop()
            
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await action.hooks.before(action, result)
                    action.setup()

                result.times['start'] = time.monotonic()

                result = await asyncio.wait_for(
                    connection.execute(action, result),
                    timeout=self.timeouts.total_timeout
                )

                result.times['complete'] = time.monotonic()
       
                self.pool.connections.append(connection)

                if action.hooks.after:
                    action = await action.hooks.after(action, result)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(result, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()    

            except Exception as e:
                result.times['complete'] = time.monotonic()
                result.error = str(e)

                self.pool.connections.append(
                    self.custom_connection(self.pool.reset_connections)
                )

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return result


from typing import Any, Callable, Awaitable
from hedra.core.graphs.simple_context import SimpleContext
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from .hook import Hook


class ContextHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        key: str=None
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            hook_type=HookType.CONTEXT
        )
        
        self.context_key = key

    async def call(self, context: SimpleContext) -> None:
        context[self.context_key] = await self._call()